# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from azure.cli.core import telemetry
from azure.cli.core.commands import AzCliCommandInvoker
from azure.cli.core.azclierror import CommandNotFoundError
from knack.log import get_logger

logger = get_logger(__name__)


def _get_extension_command_tree(cli_ctx):
    from azure.cli.core._session import EXT_CMD_TREE
    import os
    VALID_SECOND = 3600 * 24 * 10
    if not cli_ctx:
        return None
    EXT_CMD_TREE.load(os.path.join(cli_ctx.config.config_dir, 'extensionCommandTree.json'), VALID_SECOND)
    if not EXT_CMD_TREE.data:
        import posixpath
        import requests
        from azure.cli.core.util import should_disable_connection_verify
        try:
            azmirror_endpoint = cli_ctx.cloud.endpoints.azmirror_storage_account_resource_id if cli_ctx and \
                cli_ctx.cloud.endpoints.has_endpoint_set('azmirror_storage_account_resource_id') else None
            url = posixpath.join(azmirror_endpoint, 'extensions', 'extensionCommandTree.json') if \
                azmirror_endpoint else 'https://aka.ms/azExtCmdTree'
            response = requests.get(
                url,
                verify=(not should_disable_connection_verify()),
                timeout=10)
        except Exception as ex:  # pylint: disable=broad-except
            logger.info("Request failed for extension command tree: %s", str(ex))
            return None
        if response.status_code == 200:
            EXT_CMD_TREE.data = response.json()
            EXT_CMD_TREE.save_with_retry()
        else:
            logger.info("Error when retrieving extension command tree. Response code: %s", response.status_code)
            return None
    return EXT_CMD_TREE


def _get_all_extensions(cmd_chain, ext_set=None):
    """Find all the extension names in cmd_chain (dict of extension command subtree).
    An example of cmd_chain may look like (a command sub tree of the 'aks' command group):
    {
        "create": "aks-preview",
        "update": "aks-preview",
        "app": {
            "up": "deploy-to-azure"
        },
        "use-dev-spaces": "dev-spaces"
    }
    Then the resulting ext_set is {'aks-preview', 'deploy-to-azure', 'dev-spaces'}
    """
    ext_set = set() if ext_set is None else ext_set
    for key in cmd_chain:
        if isinstance(cmd_chain[key], str):
            ext_set.add(cmd_chain[key])
        else:
            _get_all_extensions(cmd_chain[key], ext_set)
    return ext_set


def _search_in_extension_commands(cli_ctx, command_str, allow_prefix_match=False):
    """Search the command in an extension commands dict which mimics a prefix tree.
    If the value of the dict item is a string, then the key represents the end of a complete command
    and the value is the name of the extension that the command belongs to.
    An example of the dict read from extensionCommandTree.json:
    {
        "aks": {
            "create": "aks-preview",
            "update": "aks-preview",
            "app": {
                "up": "deploy-to-azure"
            },
            "use-dev-spaces": "dev-spaces"
        },
        ...
    }
    """
    if not command_str:
        return None
    cmd_chain = _get_extension_command_tree(cli_ctx)
    if not cmd_chain:
        return None
    for part in command_str.split():
        try:
            if isinstance(cmd_chain[part], str):
                return cmd_chain[part]
            cmd_chain = cmd_chain[part]
        except KeyError:
            return None
    # command_str is prefix of one or more complete commands.
    if not allow_prefix_match:
        return None
    all_exts = _get_all_extensions(cmd_chain)
    return list(all_exts) if all_exts else None


def _get_extension_run_after_dynamic_install_config(cli_ctx):
    default_value = True
    run_after_extension_installed = cli_ctx.config.getboolean('extension',
                                                              'run_after_dynamic_install',
                                                              default_value) if cli_ctx else default_value
    return run_after_extension_installed


def _get_extension_allow_preview_install_config(cli_ctx):
    dynamic_install_allow_preview = None
    if cli_ctx:
        if cli_ctx.config.get('extension', 'dynamic_install_allow_preview', None) is None:
            logger.warning("Preview version of extension is disabled by default for extension installation, "
                           "enabled for modules without stable versions. ")
            logger.warning("Please run 'az config set extension.dynamic_install_allow_preview=true or false' "
                           "to config it specifically. ")
        else:
            dynamic_install_allow_preview = cli_ctx.config.getboolean('extension', 'dynamic_install_allow_preview')
    return dynamic_install_allow_preview


def try_install_extension(parser, args):
    # parser.cli_ctx is None when parser.prog is beyond 'az', such as 'az iot'.
    # use cli_ctx from cli_help which is not lost.
    cli_ctx = parser.cli_ctx or (parser.cli_help.cli_ctx if parser.cli_help else None)
    use_dynamic_install = _get_extension_use_dynamic_install_config(cli_ctx)
    if use_dynamic_install != 'no':
        # The function will exit if the parse error is caused by the extension for the command not installed
        _check_value_in_extensions(cli_ctx, parser, args, use_dynamic_install == 'yes_without_prompt')
    return use_dynamic_install


def _get_extension_use_dynamic_install_config(cli_ctx):
    from knack.prompting import verify_is_a_tty, NoTTYException
    try:
        verify_is_a_tty()
        # prompt when it's in tty
        default_value = 'yes_prompt'
    except NoTTYException:
        # without prompt when it's not in tty
        default_value = 'yes_without_prompt'
    use_dynamic_install = cli_ctx.config.get(
        'extension', 'use_dynamic_install', default_value).lower() if cli_ctx else default_value
    if use_dynamic_install not in ['no', 'yes_prompt', 'yes_without_prompt']:
        use_dynamic_install = default_value
    return use_dynamic_install


def _check_value_in_extensions(cli_ctx, parser, args, no_prompt):  # pylint: disable=too-many-statements, too-many-locals
    """Check if the command args can be found in extension commands.
       Exit command if the error is caused by an extension not installed.
       Otherwise return.
    """
    # Check if the command is from an extension
    from azure.cli.core.util import roughly_parse_command
    from azure.cli.core.azclierror import NoTTYError
    exit_code = 2
    command_str = roughly_parse_command(args[1:])
    allow_prefix_match = args[-1] == '-h' or args[-1] == '--help'
    ext_name = _search_in_extension_commands(cli_ctx, command_str, allow_prefix_match=allow_prefix_match)
    # ext_name is a list if the input command matches the prefix of one or more extension commands,
    # for instance: `az blueprint` when running `az blueprint -h`
    # ext_name is a str if the input command matches a complete command of an extension,
    # for instance: `az blueprint create`
    if isinstance(ext_name, list):
        if len(ext_name) > 1:
            from knack.prompting import prompt_choice_list, NoTTYException
            prompt_msg = "The command requires the latest version of one of the following " \
                "extensions. You need to pick one to install:"
            try:
                choice_idx = prompt_choice_list(prompt_msg, ext_name)
                ext_name = ext_name[choice_idx]
                no_prompt = True
            except NoTTYException:
                tty_err_msg = "{}{}\nUnable to prompt for selection as no tty available. Please update or " \
                    "install the extension with 'az extension add --upgrade -n <extension-name>'." \
                    .format(prompt_msg, ext_name)
                az_error = NoTTYError(tty_err_msg)
                az_error.print_error()
                az_error.send_telemetry()
                parser.exit(exit_code)
        else:
            ext_name = ext_name[0]
    if not ext_name:
        return

    # If a valid command has parser error, it may be caused by CLI running on a profile that is
    # not 'latest' and the command is not supported in that profile. If this command exists in an extension,
    # CLI will try to download the extension and rerun the command. But the parser will fail again and try to
    # install the extension and rerun the command infinitely. So we need to check if the latest version of the
    # extension is already installed and return if yes as the error is not caused by extension not installed.
    from azure.cli.core.extension import get_extension, ExtensionNotInstalledException
    from azure.cli.core.extension._resolve import resolve_from_index, NoExtensionCandidatesError
    extension_allow_preview = _get_extension_allow_preview_install_config(cli_ctx)
    try:
        ext = get_extension(ext_name)
    except ExtensionNotInstalledException:
        pass
    else:
        try:
            resolve_from_index(ext_name, cur_version=ext.version, cli_ctx=cli_ctx,
                               allow_preview=extension_allow_preview)
        except NoExtensionCandidatesError:
            return

    telemetry.set_command_details(command_str,
                                  parameters=AzCliCommandInvoker._extract_parameter_names(args),  # pylint: disable=protected-access
                                  extension_name=ext_name)
    run_after_extension_installed = _get_extension_run_after_dynamic_install_config(cli_ctx)
    prompt_info = ""
    if no_prompt:
        logger.warning('The command requires the extension %s. It will be installed first.', ext_name)
        install_ext = True
    else:  # yes_prompt
        from knack.prompting import prompt_y_n, NoTTYException
        prompt_msg = 'The command requires the extension {}. Do you want to install it now?'.format(ext_name)
        if run_after_extension_installed:
            prompt_msg = '{} The command will continue to run after the extension is installed.' \
                .format(prompt_msg)
        NO_PROMPT_CONFIG_MSG = "Run 'az config set extension.use_dynamic_install=" \
            "yes_without_prompt' to allow installing extensions without prompt."
        try:
            install_ext = prompt_y_n(prompt_msg, default='y')
            if install_ext:
                prompt_info = " with prompt"
                logger.warning(NO_PROMPT_CONFIG_MSG)
        except NoTTYException:
            tty_err_msg = "The command requires the extension {}. " \
                          "Unable to prompt for extension install confirmation as no tty " \
                          "available. {}".format(ext_name, NO_PROMPT_CONFIG_MSG)
            az_error = NoTTYError(tty_err_msg)
            az_error.print_error()
            az_error.send_telemetry()
            parser.exit(exit_code)

    print_error = True
    if install_ext:
        from azure.cli.core.extension.operations import add_extension
        add_extension(cli_ctx=cli_ctx, extension_name=ext_name, upgrade=True, allow_preview=extension_allow_preview)
        if run_after_extension_installed:
            import subprocess
            import platform
            exit_code = subprocess.call(args, shell=platform.system() == 'Windows')
            # In this case, error msg is for telemetry recording purpose only.
            # From UX perspective, the command will rerun in subprocess. Whether it succeeds or fails,
            # mesages will be shown from the subprocess and this process should not print more message to
            # interrupt that.
            print_error = False
            error_msg = ("Extension {} dynamically installed{} and commands will be "
                         "rerun automatically.").format(ext_name, prompt_info)
        else:
            error_msg = 'Extension {} installed{}. Please rerun your command.' \
                .format(ext_name, prompt_info)
    else:
        error_msg = "The command requires the latest version of extension {ext_name}. " \
            "To install, run 'az extension add --upgrade -n {ext_name}'.".format(
                ext_name=ext_name)
    az_error = CommandNotFoundError(error_msg)
    if print_error:
        az_error.print_error()
    az_error.send_telemetry()
    parser.exit(exit_code)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
import base64
import itertools
import json


class AAZPageIterator:  # pylint: disable=too-many-instance-attributes
    def __init__(self, executor, extract_result, cli_ctx, next_link, offset, limit):
        self._executor = executor
        self._extract_result = extract_result
        self._cli_ctx = cli_ctx
        self._next_link = next_link
        self._did_once_already = bool(next_link)  # desired start
        self._total = limit + offset if isinstance(limit, int) else None
        self._start = offset
        self._curr_link = None
        self._curr_size = None

    def __iter__(self):
        return self

    def __next__(self):
        def next_token(_, result):  # pylint: disable=unused-argument
            from knack.log import get_logger
            logger = get_logger(__name__)

            token = {"next_link": self._curr_link, "offset": self._curr_size + self._total}
            token = json.dumps(token).encode("utf-8")
            logger.warning("Token of next page: %s", base64.b64encode(token).decode("utf-8"))

        if self._total is not None and self._total < 0:
            from knack.events import EVENT_CLI_SUCCESSFUL_EXECUTE
            self._cli_ctx.register_event(EVENT_CLI_SUCCESSFUL_EXECUTE, next_token)

            raise StopIteration

        if not self._next_link and self._did_once_already:
            raise StopIteration("End of paginating.")

        self._executor(self._next_link)
        self._did_once_already = True
        self._curr_link = self._next_link
        curr_page, self._next_link = self._extract_result()

        if self._total is None:
            return iter(curr_page)

        start = self._start  # record 1st time only
        self._start = 0
        self._curr_size = len(curr_page)
        self._total -= self._curr_size

        if self._total < 0:
            return iter(curr_page[start:self._total])  # smaller than current page

        return iter(curr_page[start:])


class AAZPaged:
    def __init__(self, executor, extract_result, cli_ctx, token=None, limit=None):
        if isinstance(limit, int) and isinstance(token, str):
            next_token = json.loads(token)
        else:
            next_token = {"next_link": None, "offset": 0}  # default value

        self._page_iterator = itertools.chain.from_iterable(
            AAZPageIterator(
                executor=executor,
                extract_result=extract_result,
                cli_ctx=cli_ctx,
                next_link=next_token["next_link"],
                offset=next_token["offset"],
                limit=limit
            )
        )

    def __repr__(self):
        return "<iterator object azure.cli.core.aaz.paging.AAZPaged at {}>".format(hex(id(self)))

    def __iter__(self):
        return self

    def __next__(self):
        return next(self._page_iterator)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
# pylint: disable=too-few-public-methods
from knack.prompting import prompt, prompt_pass


class AAZPromptInput:

    def __init__(self, msg, help_string=None):
        self._msg = msg
        self._help_string = help_string

    def __call__(self, *args, **kwargs):
        return prompt(msg=self._msg, help_string=self._help_string)

    @property
    def help_message(self):
        return "If value is blank it's asked from the tty."


class AAZPromptPasswordInput(AAZPromptInput):

    def __init__(self, msg, confirm=False, help_string=None):
        super().__init__(msg=msg, help_string=help_string)
        self._confirm = confirm

    def __call__(self, *args, **kwargs):
        return prompt_pass(msg=self._msg, confirm=self._confirm, help_string=self._help_string)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from azure.cli.core import telemetry
from knack.log import get_logger


logger = get_logger(__name__)


class CommandRecommender:  # pylint: disable=too-few-public-methods
    """Recommend a command for user when user's command fails.
    It uses examples from help files to provide recommendations."""

    def __init__(self, command, parameters, extension, error_msg, cli_ctx):
        """
        :param command: The command name in user's input.
        :type command: str
        :param parameters: The raw parameters in users input.
        :type parameters: list
        :param extension: The extension name in user's input if the command comes from an extension.
        :type extension: str
        :param error_msg: The error message of the failed command.
        :type error_msg: str
        :param cli_ctx: CLI context when parser fails.
        :type cli_ctx: knack.cli.CLI
        """
        self.command = command.strip()
        self.parameters = parameters
        self.extension = extension
        self.error_msg = error_msg
        self.cli_ctx = cli_ctx
        # the item is a dict with the form {'command': #, 'description': #}
        self.help_examples = []

    def set_help_examples(self, examples):
        """Set help examples.

        :param examples: The examples from CLI help file.
        :type examples: list
        """

        self.help_examples.extend(examples)

    def provide_recommendations(self):
        """Provide recommendations when a command fails.

        The recommendations are from CLI help examples,
        which include both commands and reference links along with their descriptions.

        :return: The decorated recommendations
        :type: list
        """

        from azure.cli.core.style import Style, highlight_command
        from azure.cli.core.parser import OVERVIEW_REFERENCE

        def sort_recommendations(recommendations):
            """Sort the recommendations by parameter matching.

            The sorting rules below are applied in order:
                1. Commands starting with the user's input command name are ahead of those don't
                2. Commands having more matched arguments are ahead of those having less
                3. Commands having less arguments are ahead of those having more

            :param recommendations: The unordered recommendations
            :type recommendations: list
            :return: The ordered recommendations
            :type: list
            """

            candidates = []
            target_arg_list = self._normalize_parameters(self.parameters)
            for recommendation in recommendations:
                matches = 0
                arg_list = self._normalize_parameters(recommendation['command'].split(' '))

                # ignore commands that do not start with the use's input command name
                if recommendation['command'].startswith('az {}'.format(self.command)):
                    for arg in arg_list:
                        if arg in target_arg_list:
                            matches += 1
                else:
                    matches = -1

                candidates.append({
                    'recommendation': recommendation,
                    'arg_list': arg_list,
                    'matches': matches
                })

            # sort the candidates by the number of matched arguments and total arguments
            candidates.sort(key=lambda item: (item['matches'], -len(item['arg_list'])), reverse=True)

            return [candidate['recommendation'] for candidate in candidates]

        def replace_param_values(command):  # pylint: disable=unused-variable
            """Replace the parameter values in a command with user's input values

            :param command: The command whose parameter value needs to be replaced
            :type command: str
            :return: The command with parameter values being replaced
            :type: str
            """

            # replace the parameter values only when the recommended
            # command's name is the same with user's input command name
            if not command.startswith('az {}'.format(self.command)):
                return command

            source_kwargs = get_parameter_kwargs(self.parameters)
            param_mappings = self._get_param_mappings()

            return replace_parameter_values(command, source_kwargs, param_mappings)

        # do not recommend commands if it is disabled by config
        if self.cli_ctx and self.cli_ctx.config.get('core', 'error_recommendation', 'on').upper() == 'OFF':
            return []

        recommendations = self.help_examples

        # sort the recommendations by parameter matching, get the top 3 recommended commands
        recommendations = sort_recommendations(recommendations)[:3]

        raw_commands = []
        decorated_recommendations = []
        for recommendation in recommendations:
            # generate raw commands recorded in Telemetry
            raw_command = recommendation['command']
            raw_commands.append(raw_command)

            # disable the parameter replacement feature because it will make command description inaccurate
            # raw_command = replace_param_values(raw_command)

            # generate decorated commands shown to users
            decorated_command = highlight_command(raw_command)
            decorated_description = [(
                Style.SECONDARY,
                recommendation.get('description', 'No description is found.') + '\n'
            )]
            decorated_recommendations.append((decorated_command, decorated_description))

        # add reference link as a recommendation
        decorated_link = [(Style.HYPERLINK, OVERVIEW_REFERENCE)]

        decorated_description = [(Style.SECONDARY, 'Read more about the command in reference docs')]
        decorated_recommendations.append((decorated_link, decorated_description))

        # set the recommend command into Telemetry
        self._set_recommended_command_to_telemetry(raw_commands)

        return decorated_recommendations

    def _set_recommended_command_to_telemetry(self, raw_commands):
        """Set the recommended commands to Telemetry

        :param raw_commands: The recommended raw commands
        :type raw_commands: list
        """

        telemetry.set_debug_info('ExampleRecommendCommand', ';'.join(raw_commands))

    def _normalize_parameters(self, args):
        """Normalize a parameter list.

        Get the standard parameter name list of the raw parameters, which includes:
            1. Use long options to replace short options
            2. Remove the unrecognized parameter names
            3. Sort the parameter names by their lengths
        An example: ['-g', 'RG', '-n', 'NAME'] ==> ['--resource-group', '--name']

        :param args: The raw arg list of a command
        :type args: list
        :return: A standard, valid and sorted parameter name list
        :type: list
        """

        from azure.cli.core.commands import AzCliCommandInvoker

        parameters = AzCliCommandInvoker._extract_parameter_names(args)  # pylint: disable=protected-access
        normalized_parameters = []

        param_mappings = self._get_param_mappings()
        for parameter in parameters:
            if parameter in param_mappings:
                normalized_form = param_mappings.get(parameter, None) or parameter
                normalized_parameters.append(normalized_form)
            else:
                logger.debug('"%s" is an invalid parameter for command "%s".', parameter, self.command)

        return sorted(normalized_parameters)

    def _get_param_mappings(self):
        try:
            cmd_table = self.cli_ctx.invocation.commands_loader.command_table.get(self.command, None)
        except AttributeError:
            cmd_table = None

        return get_parameter_mappings(cmd_table)


def get_parameter_mappings(command_table):
    """Get the short option to long option mappings of a command

    :param parameter_table: CLI command object
    :type parameter_table: knack.commands.CLICommand
    :param command_name: The command name
    :type command name: str
    :return: The short to long option mappings of the parameters
    :type: dict
    """

    from knack.deprecation import Deprecated

    parameter_table = None
    if hasattr(command_table, 'arguments'):
        parameter_table = command_table.arguments

    param_mappings = {
        '-h': '--help',
        '-o': '--output',
        '--only-show-errors': None,
        '--help': None,
        '--output': None,
        '--query': None,
        '--debug': None,
        '--verbose': None,
        '--yes': None,
        '--no-wait': None
    }

    if parameter_table:
        for argument in parameter_table.values():
            options = argument.type.settings['options_list']
            options = [option for option in options if not isinstance(option, Deprecated)]
            # skip the positional arguments
            if not options:
                continue
            try:
                sorted_options = sorted(options, key=len, reverse=True)
                standard_form = sorted_options[0]

                for option in sorted_options[1:]:
                    param_mappings[option] = standard_form
                param_mappings[standard_form] = standard_form
            except TypeError:
                logger.debug('Unexpected argument options `%s` of type `%s`.', options, type(options).__name__)

    return param_mappings


def get_parameter_kwargs(args):
    """Get parameter name-value mappings from the raw arg list
    An example: ['-g', 'RG', '--name=NAME'] ==> {'-g': 'RG', '--name': 'NAME'}

    :param args: The raw arg list of a command
    :type args: list
    :return: The parameter name-value mappings
    :type: dict
    """

    parameter_kwargs = {}
    for index, parameter in enumerate(args):
        if parameter.startswith('-'):

            param_name, param_val = parameter, None
            if '=' in parameter:
                pieces = parameter.split('=')
                param_name, param_val = pieces[0], pieces[1]
            elif index + 1 < len(args) and not args[index + 1].startswith('-'):
                param_val = args[index + 1]

            if param_val is not None and ' ' in param_val:
                param_val = '"{}"'.format(param_val)
            parameter_kwargs[param_name] = param_val

    return parameter_kwargs


def replace_parameter_values(target_command, source_kwargs, param_mappings):
    """Replace the parameter values in target_command with values in source_kwargs

    :param target_command: The command in which the parameter values need to be replaced
    :type target_command: str
    :param source_kwargs: The source key-val pairs used to replace the values
    :type source_kwargs: dict
    :param param_mappings: The short-long option mappings in terms of the target_command
    :type param_mappings: dict
    :returns: The target command with parameter values being replaced
    :type: str
    """

    def get_user_param_value(target_param):
        """Get the value that is used as the replaced value of target_param

        :param target_param: The parameter name whose value needs to be replaced
        :type target_param: str
        :return: The replaced value for target_param
        :type: str
        """
        standard_source_kwargs = {}

        for param, val in source_kwargs.items():
            if param in param_mappings:
                standard_param = param_mappings[param]
                standard_source_kwargs[standard_param] = val

        if target_param in param_mappings:
            standard_target_param = param_mappings[target_param]
            if standard_target_param in standard_source_kwargs:
                return standard_source_kwargs[standard_target_param]

        return None

    command_args = target_command.split(' ')
    for index, arg in enumerate(command_args):
        if arg.startswith('-') and index + 1 < len(command_args) and not command_args[index + 1].startswith('-'):
            user_param_val = get_user_param_value(arg)
            if user_param_val:
                command_args[index + 1] = user_param_val

    return ' '.join(command_args)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

# This file is modified from
# https://github.com/AzureAD/microsoft-authentication-extensions-for-python/blob/dev/sample/token_cache_sample.py

import json
import sys

from msal_extensions import (FilePersistenceWithDataProtection, KeychainPersistence, LibsecretPersistence,
                             FilePersistence, PersistedTokenCache, CrossPlatLock)
from msal_extensions.persistence import PersistenceNotFound

from knack.log import get_logger
from azure.cli.core.decorators import retry


logger = get_logger(__name__)

# Files extensions for encrypted and plaintext persistence
file_extensions = {True: '.bin', False: '.json'}


def load_persisted_token_cache(location, encrypt):
    persistence = build_persistence(location, encrypt)
    return PersistedTokenCache(persistence)


def load_secret_store(location, encrypt):
    persistence = build_persistence(location, encrypt)
    return SecretStore(persistence)


def build_persistence(location, encrypt):
    """Build a suitable persistence instance based your current OS"""
    location += file_extensions[encrypt]
    logger.debug("build_persistence: location=%r, encrypt=%r", location, encrypt)
    if encrypt:
        if sys.platform.startswith('win'):
            return FilePersistenceWithDataProtection(location)
        if sys.platform.startswith('darwin'):
            return KeychainPersistence(location, "my_service_name", "my_account_name")
        if sys.platform.startswith('linux'):
            return LibsecretPersistence(
                location,
                schema_name="my_schema_name",
                attributes={"my_attr1": "foo", "my_attr2": "bar"}
            )
    else:
        return FilePersistence(location)


class SecretStore:
    def __init__(self, persistence):
        self._lock_file = persistence.get_location() + ".lockfile"
        self._persistence = persistence

    def save(self, content):
        with CrossPlatLock(self._lock_file):
            self._persistence.save(json.dumps(content, indent=4))

    @retry()
    def load(self):
        try:
            return json.loads(self._persistence.load())
        except PersistenceNotFound:
            return []



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
# pylint: disable=line-too-long

import os
import json
from pprint import pformat
import configparser

from azure.cli.core.profiles import API_PROFILES
from azure.cli.core._config import GLOBAL_CONFIG_DIR
from azure.cli.core.util import urlretrieve

from knack.log import get_logger
from knack.util import CLIError

logger = get_logger(__name__)

CLOUD_CONFIG_FILE = os.path.join(GLOBAL_CONFIG_DIR, 'clouds.config')

# Add names of clouds that don't allow telemetry data collection here such as some air-gapped clouds.
CLOUDS_FORBIDDING_TELEMETRY = ['USSec', 'USNat']


class CloudNotRegisteredException(Exception):
    def __init__(self, cloud_name):
        super().__init__(cloud_name)
        self.cloud_name = cloud_name

    def __str__(self):
        return "The cloud '{}' is not registered.".format(self.cloud_name)


class CloudAlreadyRegisteredException(Exception):
    def __init__(self, cloud_name):
        super().__init__(cloud_name)
        self.cloud_name = cloud_name

    def __str__(self):
        return "The cloud '{}' is already registered.".format(self.cloud_name)


class CannotUnregisterCloudException(Exception):
    pass


class CloudEndpointNotSetException(CLIError):
    pass


class CloudSuffixNotSetException(CLIError):
    pass


class CloudEndpoints:  # pylint: disable=too-few-public-methods,too-many-instance-attributes

    ARM_METADATA_INDEX = {
        "active_directory": "authentication.loginEndpoint",
        "active_directory_data_lake_resource_id": "activeDirectoryDataLake",
        "active_directory_graph_resource_id": "graphAudience",
        "active_directory_resource_id": "authentication.audiences[0]",
        "app_insights_resource_id": "appInsightsResourceId",
        "app_insights_telemetry_channel_resource_id": "appInsightsTelemetryChannelResourceId",
        "app_service_resource_id": "appServiceResourceId",
        "attestation_resource_id": "attestationResourceId",
        "azmirror_storage_account_resource_id": "azmirrorStorageAccountResourceId",
        "batch_resource_id": "batch",
        "gallery": "gallery",
        "log_analytics_resource_id": "logAnalyticsResourceId",
        "management": "authentication.audiences[0]",
        "media_resource_id": "media",
        "microsoft_graph_resource_id": "microsoftGraphResourceId",
        "ossrdbms_resource_id": "ossrdbmsResourceId",
        "portal": "portal",
        "resource_manager": "resourceManager",
        "sql_management": "sqlManagement",
        "synapse_analytics_resource_id": "synapseAnalyticsResourceId",
        "vm_image_alias_doc": "vmImageAliasDoc",
    }  # Please keep the endpoints in alphabetical order

    def __init__(self,  # pylint: disable=unused-argument
                 active_directory=None,
                 active_directory_data_lake_resource_id=None,
                 active_directory_graph_resource_id=None,
                 active_directory_resource_id=None,
                 app_insights_resource_id=None,
                 app_insights_telemetry_channel_resource_id=None,
                 app_service_resource_id=None,
                 attestation_resource_id=None,
                 azmirror_storage_account_resource_id=None,
                 batch_resource_id=None,
                 gallery=None,
                 log_analytics_resource_id=None,
                 management=None,
                 media_resource_id=None,
                 microsoft_graph_resource_id=None,
                 ossrdbms_resource_id=None,
                 portal=None,
                 resource_manager=None,
                 sql_management=None,
                 synapse_analytics_resource_id=None,
                 vm_image_alias_doc=None,
                 **kwargs):  # To support init with __dict__ for deserialization
        # Attribute names are significant. They are used when storing/retrieving clouds from config
        self.active_directory = active_directory
        self.active_directory_data_lake_resource_id = active_directory_data_lake_resource_id
        self.active_directory_graph_resource_id = active_directory_graph_resource_id
        self.active_directory_resource_id = active_directory_resource_id
        self.app_insights_resource_id = app_insights_resource_id
        self.app_insights_telemetry_channel_resource_id = app_insights_telemetry_channel_resource_id
        self.app_service_resource_id = app_service_resource_id
        self.attestation_resource_id = attestation_resource_id
        self.azmirror_storage_account_resource_id = azmirror_storage_account_resource_id
        self.batch_resource_id = batch_resource_id
        self.gallery = gallery
        self.log_analytics_resource_id = log_analytics_resource_id
        self.management = management
        self.media_resource_id = media_resource_id
        self.microsoft_graph_resource_id = microsoft_graph_resource_id
        self.ossrdbms_resource_id = ossrdbms_resource_id
        self.portal = portal
        self.resource_manager = resource_manager
        self.sql_management = sql_management
        self.synapse_analytics_resource_id = synapse_analytics_resource_id
        self.vm_image_alias_doc = vm_image_alias_doc
        # Please keep the endpoints in alphabetical order

    def has_endpoint_set(self, endpoint_name):
        try:
            # Can't simply use hasattr here as we override __getattribute__ below.
            # Python 3 hasattr() only returns False if an AttributeError is raised but we raise
            # CloudEndpointNotSetException. This exception is not a subclass of AttributeError.
            getattr(self, endpoint_name)
            return True
        except Exception:  # pylint: disable=broad-except
            return False

    def __getattribute__(self, name):
        val = object.__getattribute__(self, name)
        if val is None:
            raise CloudEndpointNotSetException("The endpoint '{}' for this cloud "
                                               "is not set but is used.\n"
                                               "{} may be corrupt or invalid.\nResolve the error or delete this file "
                                               "and try again.".format(name, CLOUD_CONFIG_FILE))
        return val


class CloudSuffixes:  # pylint: disable=too-few-public-methods,too-many-instance-attributes

    ARM_METADATA_INDEX = {
        "acr_login_server_endpoint": "suffixes.acrLoginServer",
        "attestation_endpoint": "suffixes.attestationEndpoint",
        "azure_datalake_analytics_catalog_and_job_endpoint": "suffixes.azureDataLakeAnalyticsCatalogAndJob",
        "azure_datalake_store_file_system_endpoint": "suffixes.azureDataLakeStoreFileSystem",
        "keyvault_dns": "suffixes.keyVaultDns",
        "mariadb_server_endpoint": "suffixes.mariadbServerEndpoint",
        "mhsm_dns": "suffixes.mhsmDns",
        "mysql_server_endpoint": "suffixes.mysqlServerEndpoint",
        "postgresql_server_endpoint": "suffixes.postgresqlServerEndpoint",
        "sql_server_hostname": "suffixes.sqlServerHostname",
        "storage_endpoint": "suffixes.storage",
        "storage_sync_endpoint": "suffixes.storageSyncEndpointSuffix",
        "synapse_analytics_endpoint": "suffixes.synapseAnalytics"
    }  # Please keep the suffixes in alphabetical order

    def __init__(self,  # pylint: disable=unused-argument
                 acr_login_server_endpoint=None,
                 attestation_endpoint=None,
                 azure_datalake_analytics_catalog_and_job_endpoint=None,
                 azure_datalake_store_file_system_endpoint=None,
                 keyvault_dns=None,
                 mariadb_server_endpoint=None,
                 mhsm_dns=None,
                 mysql_server_endpoint=None,
                 postgresql_server_endpoint=None,
                 sql_server_hostname=None,
                 storage_endpoint=None,
                 storage_sync_endpoint=None,
                 synapse_analytics_endpoint=None,
                 **kwargs):  # To support init with __dict__ for deserialization
        # Attribute names are significant. They are used when storing/retrieving clouds from config
        self.acr_login_server_endpoint = acr_login_server_endpoint
        self.attestation_endpoint = attestation_endpoint
        self.azure_datalake_analytics_catalog_and_job_endpoint = azure_datalake_analytics_catalog_and_job_endpoint
        self.azure_datalake_store_file_system_endpoint = azure_datalake_store_file_system_endpoint
        self.keyvault_dns = keyvault_dns
        self.mariadb_server_endpoint = mariadb_server_endpoint
        self.mhsm_dns = mhsm_dns
        self.mysql_server_endpoint = mysql_server_endpoint
        self.postgresql_server_endpoint = postgresql_server_endpoint
        self.sql_server_hostname = sql_server_hostname
        self.storage_endpoint = storage_endpoint
        self.storage_sync_endpoint = storage_sync_endpoint
        self.synapse_analytics_endpoint = synapse_analytics_endpoint
        # Please keep the suffixes in alphabetical order

    def __getattribute__(self, name):
        val = object.__getattribute__(self, name)
        if val is None:
            raise CloudSuffixNotSetException("The suffix '{}' for this cloud "
                                             "is not set but is used.\n"
                                             "{} may be corrupt or invalid.\nResolve the error or delete this file "
                                             "and try again.".format(name, CLOUD_CONFIG_FILE))
        return val


def _get_database_server_endpoint(sql_server_hostname, cloud_name):
    def _concat_db_server_endpoint(db_prefix):
        if cloud_name == 'AzureCloud':
            return db_prefix + '.database.azure.com'
        if not sql_server_hostname:
            return None
        return db_prefix + sql_server_hostname
    return _concat_db_server_endpoint


def _get_endpoint_fallback_value(cloud_name):
    def _get_cloud_endpoint_fallback_value(endpoint_name):
        endpoint_mapper = {c.name: c.endpoints.__dict__.get(endpoint_name, None) for c in HARD_CODED_CLOUD_LIST}
        return endpoint_mapper.get(cloud_name, None)
    return _get_cloud_endpoint_fallback_value


def _get_suffix_fallback_value(cloud_name):
    def _get_cloud_suffix_fallback_value(suffix_name):
        suffix_mapper = {c.name: c.suffixes.__dict__.get(suffix_name, None) for c in HARD_CODED_CLOUD_LIST}
        return suffix_mapper.get(cloud_name, None)
    return _get_cloud_suffix_fallback_value


def _convert_arm_to_cli(arm_cloud_metadata_dict):
    cli_cloud_metadata_dict = {}
    for cloud in arm_cloud_metadata_dict:
        cli_cloud_metadata_dict[cloud['name']] = _arm_to_cli_mapper(cloud)
    if 'AzureCloud' in cli_cloud_metadata_dict:
        cli_cloud_metadata_dict['AzureCloud'].endpoints.active_directory = 'https://login.microsoftonline.com'  # change once active_directory is fixed in ARM for the public cloud
    return cli_cloud_metadata_dict


def _add_starting_dot(suffix):
    return suffix if not suffix or suffix.startswith('.') else '.' + suffix


def _get_arm_endpoint(arm_dict, is_suffix=False):
    def _get_processed_arm_endpoint(name, add_dot=False, fallback_value=None):
        if is_suffix:
            return (_add_starting_dot(arm_dict['suffixes'][name]) if add_dot else arm_dict['suffixes'][name]) if name in arm_dict['suffixes'] else fallback_value
        return arm_dict[name] if name in arm_dict else fallback_value
    return _get_processed_arm_endpoint


def _arm_to_cli_mapper(arm_dict):
    get_endpoint = _get_arm_endpoint(arm_dict)
    get_suffix = _get_arm_endpoint(arm_dict, is_suffix=True)

    sql_server_hostname = get_suffix('sqlServerHostname', add_dot=True)
    get_db_server_endpoint = _get_database_server_endpoint(sql_server_hostname, arm_dict['name'])

    get_suffix_fallback_value = _get_suffix_fallback_value(arm_dict['name'])
    get_endpoint_fallback_value = _get_endpoint_fallback_value(arm_dict['name'])

    return Cloud(
        arm_dict['name'],
        endpoints=CloudEndpoints(  # please add fallback_value if the endpoint is not added to https://management.azure.com/metadata/endpoints?api-version=2019-05-01 yet
            active_directory=arm_dict['authentication']['loginEndpoint'],
            active_directory_data_lake_resource_id=get_endpoint('activeDirectoryDataLake'),
            active_directory_graph_resource_id=get_endpoint('graphAudience'),
            active_directory_resource_id=arm_dict['authentication']['audiences'][0],
            app_insights_resource_id=get_endpoint('appInsightsResourceId',
                                                  fallback_value=get_endpoint_fallback_value('app_insights_resource_id')),
            app_insights_telemetry_channel_resource_id=get_endpoint('appInsightsTelemetryChannelResourceId',
                                                                    fallback_value=get_endpoint_fallback_value('app_insights_telemetry_channel_resource_id')),
            app_service_resource_id=get_endpoint('appServiceResourceId', fallback_value=get_endpoint_fallback_value('app_service_resource_id')),
            attestation_resource_id=get_endpoint('attestationResourceId',
                                                 fallback_value=get_endpoint_fallback_value('attestation_resource_id')),
            azmirror_storage_account_resource_id=get_endpoint('azmirrorStorageAccountResourceId'),
            batch_resource_id=get_endpoint('batch'),
            gallery=get_endpoint('gallery'),
            log_analytics_resource_id=get_endpoint('logAnalyticsResourceId',
                                                   fallback_value=get_endpoint_fallback_value('log_analytics_resource_id')),
            management=arm_dict['authentication']['audiences'][0],
            media_resource_id=get_endpoint('media'),
            microsoft_graph_resource_id=get_endpoint('microsoftGraphResourceId',
                                                     fallback_value=get_endpoint_fallback_value('microsoft_graph_resource_id')),  # change once microsoft_graph_resource_id is fixed in ARM
            ossrdbms_resource_id=get_endpoint('ossrdbmsResourceId',
                                              fallback_value=get_endpoint_fallback_value('ossrdbms_resource_id')),  # change once ossrdbms_resource_id is available via ARM
            portal=get_endpoint('portal'),
            resource_manager=get_endpoint('resourceManager'),
            sql_management=get_endpoint('sqlManagement'),
            synapse_analytics_resource_id=get_endpoint('synapseAnalyticsResourceId',
                                                       fallback_value=get_endpoint_fallback_value('synapse_analytics_resource_id')),
            vm_image_alias_doc=get_endpoint('vmImageAliasDoc'),
        ),  # Please keep the endpoints in alphabetical order
        suffixes=CloudSuffixes(
            acr_login_server_endpoint=get_suffix('acrLoginServer', add_dot=True),
            attestation_endpoint=get_suffix('attestationEndpoint', add_dot=True,
                                            fallback_value=get_suffix_fallback_value('attestation_endpoint')),
            azure_datalake_analytics_catalog_and_job_endpoint=get_suffix('azureDataLakeAnalyticsCatalogAndJob'),
            azure_datalake_store_file_system_endpoint=get_suffix('azureDataLakeStoreFileSystem'),
            keyvault_dns=get_suffix('keyVaultDns', add_dot=True),
            mariadb_server_endpoint=get_suffix('mariadbServerEndpoint', add_dot=True,
                                               fallback_value=get_db_server_endpoint('.mariadb')),
            mhsm_dns=get_suffix('mhsmDns', add_dot=True, fallback_value=get_suffix_fallback_value('mhsm_dns')),
            mysql_server_endpoint=get_suffix('mysqlServerEndpoint', add_dot=True,
                                             fallback_value=get_db_server_endpoint('.mysql')),
            postgresql_server_endpoint=get_suffix('postgresqlServerEndpoint', add_dot=True,
                                                  fallback_value=get_db_server_endpoint('.postgres')),
            sql_server_hostname=sql_server_hostname,
            storage_endpoint=get_suffix('storage'),
            storage_sync_endpoint=get_suffix('storageSyncEndpointSuffix', fallback_value=get_suffix_fallback_value('storage_sync_endpoint')),
            synapse_analytics_endpoint=get_suffix('synapseAnalytics', add_dot=True, fallback_value=get_suffix_fallback_value('synapse_analytics_endpoint')),
        )  # Please keep the suffixes in alphabetical order
    )


class Cloud:  # pylint: disable=too-few-public-methods
    """ Represents an Azure Cloud instance """

    def __init__(self,
                 name,
                 endpoints=None,
                 suffixes=None,
                 profile=None,
                 is_active=False):
        self.name = name
        self.endpoints = endpoints or CloudEndpoints()
        self.suffixes = suffixes or CloudSuffixes()
        self.profile = profile
        self.is_active = is_active

    def __str__(self):
        o = {
            'profile': self.profile,
            'name': self.name,
            'is_active': self.is_active,
            'endpoints': vars(self.endpoints),
            'suffixes': vars(self.suffixes),
        }
        return pformat(o)

    def to_json(self):
        return {'name': self.name, "endpoints": self.endpoints.__dict__, "suffixes": self.suffixes.__dict__}

    @classmethod
    def from_json(cls, json_str):
        return cls(json_str['name'],
                   endpoints=CloudEndpoints(**json_str['endpoints']),
                   suffixes=CloudSuffixes(**json_str['suffixes']))


class CloudNameEnum:  # pylint: disable=too-few-public-methods
    AzureCloud = 'AzureCloud'
    AzureChinaCloud = 'AzureChinaCloud'
    AzureUSGovernment = 'AzureUSGovernment'
    AzureGermanCloud = 'AzureGermanCloud'
    AzureBleuCloud = 'AzureBleuCloud'


AZURE_PUBLIC_CLOUD = Cloud(
    CloudNameEnum.AzureCloud,
    endpoints=CloudEndpoints(
        management='https://management.core.windows.net/',
        resource_manager='https://management.azure.com/',
        sql_management='https://management.core.windows.net:8443/',
        batch_resource_id='https://batch.core.windows.net/',
        gallery='https://gallery.azure.com/',
        active_directory='https://login.microsoftonline.com',
        active_directory_resource_id='https://management.core.windows.net/',
        active_directory_graph_resource_id='https://graph.windows.net/',
        microsoft_graph_resource_id='https://graph.microsoft.com/',
        active_directory_data_lake_resource_id='https://datalake.azure.net/',
        vm_image_alias_doc='https://azcliprod.blob.core.windows.net/cli/vm/aliases.json',
        media_resource_id='https://rest.media.azure.net',
        ossrdbms_resource_id='https://ossrdbms-aad.database.windows.net',
        app_insights_resource_id='https://api.applicationinsights.io',
        log_analytics_resource_id='https://api.loganalytics.io',
        app_insights_telemetry_channel_resource_id='https://dc.applicationinsights.azure.com/v2/track',
        synapse_analytics_resource_id='https://dev.azuresynapse.net',
        attestation_resource_id='https://attest.azure.net',
        app_service_resource_id='https://appservice.azure.com',
        portal='https://portal.azure.com'),
    suffixes=CloudSuffixes(
        storage_endpoint='core.windows.net',
        storage_sync_endpoint='afs.azure.net',
        keyvault_dns='.vault.azure.net',
        mhsm_dns='.managedhsm.azure.net',
        sql_server_hostname='.database.windows.net',
        mysql_server_endpoint='.mysql.database.azure.com',
        postgresql_server_endpoint='.postgres.database.azure.com',
        mariadb_server_endpoint='.mariadb.database.azure.com',
        azure_datalake_store_file_system_endpoint='azuredatalakestore.net',
        azure_datalake_analytics_catalog_and_job_endpoint='azuredatalakeanalytics.net',
        acr_login_server_endpoint='.azurecr.io',
        synapse_analytics_endpoint='.dev.azuresynapse.net',
        attestation_endpoint='.attest.azure.net'))

AZURE_CHINA_CLOUD = Cloud(
    CloudNameEnum.AzureChinaCloud,
    endpoints=CloudEndpoints(
        management='https://management.core.chinacloudapi.cn/',
        resource_manager='https://management.chinacloudapi.cn',
        sql_management='https://management.core.chinacloudapi.cn:8443/',
        batch_resource_id='https://batch.chinacloudapi.cn/',
        gallery='https://gallery.chinacloudapi.cn/',
        active_directory='https://login.chinacloudapi.cn',
        active_directory_resource_id='https://management.core.chinacloudapi.cn/',
        active_directory_graph_resource_id='https://graph.chinacloudapi.cn/',
        microsoft_graph_resource_id='https://microsoftgraph.chinacloudapi.cn',
        vm_image_alias_doc='https://azcliprod.blob.core.windows.net/cli/vm/aliases.json',
        media_resource_id='https://rest.media.chinacloudapi.cn',
        ossrdbms_resource_id='https://ossrdbms-aad.database.chinacloudapi.cn',
        app_insights_resource_id='https://api.applicationinsights.azure.cn',
        log_analytics_resource_id='https://api.loganalytics.azure.cn',
        app_insights_telemetry_channel_resource_id='https://dc.applicationinsights.azure.cn/v2/track',
        synapse_analytics_resource_id='https://dev.azuresynapse.azure.cn',
        # App Service Audience is currently not available in Mooncake,
        # Using the management endpoint until the App Service audience is enabled
        app_service_resource_id='https://management.core.chinacloudapi.cn/',
        portal='https://portal.azure.cn'),
    suffixes=CloudSuffixes(
        storage_endpoint='core.chinacloudapi.cn',
        keyvault_dns='.vault.azure.cn',
        mhsm_dns='.managedhsm.azure.cn',
        sql_server_hostname='.database.chinacloudapi.cn',
        mysql_server_endpoint='.mysql.database.chinacloudapi.cn',
        postgresql_server_endpoint='.postgres.database.chinacloudapi.cn',
        mariadb_server_endpoint='.mariadb.database.chinacloudapi.cn',
        acr_login_server_endpoint='.azurecr.cn',
        synapse_analytics_endpoint='.dev.azuresynapse.azure.cn'))

AZURE_US_GOV_CLOUD = Cloud(
    CloudNameEnum.AzureUSGovernment,
    endpoints=CloudEndpoints(
        management='https://management.core.usgovcloudapi.net/',
        resource_manager='https://management.usgovcloudapi.net/',
        sql_management='https://management.core.usgovcloudapi.net:8443/',
        batch_resource_id='https://batch.core.usgovcloudapi.net/',
        gallery='https://gallery.usgovcloudapi.net/',
        active_directory='https://login.microsoftonline.us',
        active_directory_resource_id='https://management.core.usgovcloudapi.net/',
        active_directory_graph_resource_id='https://graph.microsoftazure.us/',
        microsoft_graph_resource_id='https://graph.microsoft.us/',
        vm_image_alias_doc='https://azcliprod.blob.core.windows.net/cli/vm/aliases.json',
        media_resource_id='https://rest.media.usgovcloudapi.net',
        ossrdbms_resource_id='https://ossrdbms-aad.database.usgovcloudapi.net',
        app_insights_resource_id='https://api.applicationinsights.us',
        log_analytics_resource_id='https://api.loganalytics.us',
        app_insights_telemetry_channel_resource_id='https://dc.applicationinsights.us/v2/track',
        synapse_analytics_resource_id='https://dev.azuresynapse.usgovcloudapi.net',
        # App Service Audience is currently not available in US Government cloud,
        # Using the management endpoint until the App Service audience is enabled
        app_service_resource_id='https://management.core.usgovcloudapi.net/',
        portal='https://portal.azure.us'),
    suffixes=CloudSuffixes(
        storage_endpoint='core.usgovcloudapi.net',
        storage_sync_endpoint='afs.azure.us',
        keyvault_dns='.vault.usgovcloudapi.net',
        mhsm_dns='.managedhsm.usgovcloudapi.net',
        sql_server_hostname='.database.usgovcloudapi.net',
        mysql_server_endpoint='.mysql.database.usgovcloudapi.net',
        postgresql_server_endpoint='.postgres.database.usgovcloudapi.net',
        mariadb_server_endpoint='.mariadb.database.usgovcloudapi.net',
        acr_login_server_endpoint='.azurecr.us',
        synapse_analytics_endpoint='.dev.azuresynapse.usgovcloudapi.net'))

AZURE_GERMAN_CLOUD = Cloud(
    CloudNameEnum.AzureGermanCloud,
    endpoints=CloudEndpoints(
        management='https://management.core.cloudapi.de/',
        resource_manager='https://management.microsoftazure.de',
        sql_management='https://management.core.cloudapi.de:8443/',
        batch_resource_id='https://batch.cloudapi.de/',
        gallery='https://gallery.cloudapi.de/',
        active_directory='https://login.microsoftonline.de',
        active_directory_resource_id='https://management.core.cloudapi.de/',
        active_directory_graph_resource_id='https://graph.cloudapi.de/',
        microsoft_graph_resource_id='https://graph.microsoft.de',
        vm_image_alias_doc='https://azcliprod.blob.core.windows.net/cli/vm/aliases.json',
        media_resource_id='https://rest.media.cloudapi.de',
        ossrdbms_resource_id='https://ossrdbms-aad.database.cloudapi.de',
        portal='https://portal.microsoftazure.de'),
    suffixes=CloudSuffixes(
        storage_endpoint='core.cloudapi.de',
        keyvault_dns='.vault.microsoftazure.de',
        mhsm_dns='.managedhsm.microsoftazure.de',
        sql_server_hostname='.database.cloudapi.de',
        mysql_server_endpoint='.mysql.database.cloudapi.de',
        postgresql_server_endpoint='.postgres.database.cloudapi.de',
        mariadb_server_endpoint='.mariadb.database.cloudapi.de'))

AZURE_BLEU_CLOUD = Cloud(
    CloudNameEnum.AzureBleuCloud,
    endpoints=CloudEndpoints(
        management='https://management.sovcloud-api.fr/',
        resource_manager='https://management.sovcloud-api.fr',
        sql_management='https://management.database.sovcloud-api.fr:8443/',
        batch_resource_id='https://batch.sovcloud-api.fr/',
        gallery='https://gallery.sovcloud-api.fr/',
        active_directory='https://login.sovcloud-identity.fr',
        active_directory_resource_id='https://management.sovcloud-api.fr/',
        active_directory_graph_resource_id='https://graph.svc.sovcloud.fr/',
        microsoft_graph_resource_id='https://graph.svc.sovcloud.fr',
        vm_image_alias_doc='https://azcliprod.blob.core.windows.net/cli/vm/aliases_master.json',
        media_resource_id='https://rest.media.sovcloud-api.fr',
        ossrdbms_resource_id='https://ossrdbms-aad.database.sovcloud-api.fr',
        portal='https://portal.sovcloud-azure.fr'),
    suffixes=CloudSuffixes(
        acr_login_server_endpoint='.azurecr.sovcloud-azure.fr',
        attestation_endpoint='attest.sovcloud-api.fr',
        storage_endpoint='core.sovcloud-api.fr',
        storage_sync_endpoint='afs.sovcloud-api.fr',
        keyvault_dns='.vault.sovcloud-api.fr',
        mhsm_dns='.managedhsm.sovcloud-api.fr',
        sql_server_hostname='.database.sovcloud-api.fr',
        mysql_server_endpoint='.mysql.database.sovcloud-api.fr',
        postgresql_server_endpoint='.postgres.database.sovcloud-api.fr',
        mariadb_server_endpoint='.mariadb.database.sovcloud-api.fr',
        synapse_analytics_endpoint='.dev.azuresynapse.sovcloud-api.fr'))

HARD_CODED_CLOUD_LIST = [AZURE_PUBLIC_CLOUD, AZURE_CHINA_CLOUD, AZURE_US_GOV_CLOUD, AZURE_GERMAN_CLOUD, AZURE_BLEU_CLOUD]


def retrieve_arm_cloud_metadata():
    """ Retrieve Cloud metadata from ARM endpoint api defined in ARM_CLOUD_METADATA_URL"""
    if 'ARM_CLOUD_METADATA_URL' in os.environ:
        return json.loads(urlretrieve(os.getenv('ARM_CLOUD_METADATA_URL')))


def get_known_clouds(refresh=False):
    if 'ARM_CLOUD_METADATA_URL' in os.environ:
        from azure.cli.core._session import CLOUD_ENDPOINTS
        if not os.path.isdir(GLOBAL_CONFIG_DIR):
            os.makedirs(GLOBAL_CONFIG_DIR)
        endpoints_file = os.path.join(GLOBAL_CONFIG_DIR, 'cloudEndpoints.json')
        CLOUD_ENDPOINTS.load(endpoints_file)
        if refresh:
            CLOUD_ENDPOINTS['clouds'] = {}
        clouds = []
        if CLOUD_ENDPOINTS['clouds']:
            try:
                clouds = [Cloud.from_json(c) for c in CLOUD_ENDPOINTS['clouds']]
                logger.info("Cloud endpoints loaded from local file: %s", endpoints_file)
            except Exception as ex:  # pylint: disable=broad-except
                logger.info("Failed to parse cloud endpoints from local file. CLI will clean it and reload from ARM_CLOUD_METADATA_URL. %s", str(ex))
                CLOUD_ENDPOINTS['clouds'] = {}

        if not CLOUD_ENDPOINTS['clouds']:
            try:
                arm_cloud_dict = retrieve_arm_cloud_metadata()
                cli_cloud_dict = _convert_arm_to_cli(arm_cloud_dict)
                if 'AzureCloud' in cli_cloud_dict:
                    cli_cloud_dict['AzureCloud'].endpoints.active_directory = 'https://login.microsoftonline.com'  # change once active_directory is fixed in ARM for the public cloud
                clouds = list(cli_cloud_dict.values())
                CLOUD_ENDPOINTS['clouds'] = [c.to_json() for c in clouds]
                logger.info("Cloud endpoints loaded from ARM_CLOUD_METADATA_URL: %s", os.getenv('ARM_CLOUD_METADATA_URL'))
            except Exception as ex:  # pylint: disable=broad-except
                logger.warning('Failed to load cloud metadata from the url specified by ARM_CLOUD_METADATA_URL')
                raise ex
        if not clouds:
            raise CLIError("No clouds available. Please ensure ARM_CLOUD_METADATA_URL is valid.")
        return clouds
    return HARD_CODED_CLOUD_LIST


KNOWN_CLOUDS = get_known_clouds()


def refresh_known_clouds():
    global KNOWN_CLOUDS  # pylint:disable=global-statement
    KNOWN_CLOUDS = get_known_clouds(refresh=True)


def _set_active_cloud(cli_ctx, cloud_name):
    cli_ctx.config.set_value('cloud', 'name', cloud_name)
    cli_ctx.cloud = get_cloud(cli_ctx, cloud_name)


def get_active_cloud_name(cli_ctx):
    try:
        return cli_ctx.config.get('cloud', 'name')
    except (configparser.NoOptionError, configparser.NoSectionError):
        default_cloud_name = get_default_cloud_name()
        _set_active_cloud(cli_ctx, default_cloud_name)
        return default_cloud_name


def get_default_cloud_name():
    """ Pick AzureCloud as the default cloud if it is available, otherwise pick the first in the list"""
    if CloudNameEnum.AzureCloud.lower() in [c.name.lower() for c in KNOWN_CLOUDS]:
        return CloudNameEnum.AzureCloud
    return KNOWN_CLOUDS[0].name


def _get_cloud(cli_ctx, cloud_name):
    return next((x for x in get_clouds(cli_ctx) if x.name == cloud_name), None)


def cloud_is_registered(cli_ctx, cloud_name):
    return bool(_get_cloud(cli_ctx, cloud_name))


def get_custom_clouds(cli_ctx):
    known_cloud_names = [c.name for c in KNOWN_CLOUDS]
    return [c for c in get_clouds(cli_ctx) if c.name not in known_cloud_names]


def _get_cloud_name(cli_ctx, cloud_name):
    return next((x.name for x in get_clouds(cli_ctx) if x.name.lower() == cloud_name.lower()), cloud_name)


def get_clouds(cli_ctx):
    clouds = []
    config = configparser.ConfigParser()
    # Start off with known clouds and apply config file on top of current config
    for c in KNOWN_CLOUDS:
        _config_add_cloud(config, c)
    try:
        config.read(CLOUD_CONFIG_FILE)
    except configparser.MissingSectionHeaderError:
        os.remove(CLOUD_CONFIG_FILE)
        logger.warning("'%s' is in bad format and has been removed.", CLOUD_CONFIG_FILE)
    active_cloud_name = get_active_cloud_name(cli_ctx)
    for section in config.sections():
        c = Cloud(section)
        for option in config.options(section):
            if option == 'profile':
                c.profile = config.get(section, option)
            if option.startswith('endpoint_'):
                setattr(c.endpoints, option.replace('endpoint_', ''), config.get(section, option))
            elif option.startswith('suffix_'):
                setattr(c.suffixes, option.replace('suffix_', ''), config.get(section, option))
        if c.profile is None:
            # If profile isn't set, use latest
            setattr(c, 'profile', 'latest')
        if c.profile not in API_PROFILES:
            if c.profile in (
                "2017-03-09-profile",
                "2018-03-01-hybrid",
                "2019-03-01-hybrid",
                "2020-09-01-hybrid",
            ):
                if c.name == active_cloud_name:
                    # only apply to the active cloud
                    logger.error(
                        "The azure stack profile '%s' has been deprecated and removed, using the 'latest' profile instead.\n"
                        "To continue using Azure Stack, please install the Azure CLI `2.66.*` (LTS) version. For more details, refer to: https://learn.microsoft.com/en-us/cli/azure/whats-new-overview#important-notice-for-azure-stack-hub-customers", c.profile
                    )
                    c.profile = 'latest'
            else:
                raise CLIError('Profile {} does not exist or is not supported.'.format(c.profile))
        if not c.endpoints.has_endpoint_set('management') and \
                c.endpoints.has_endpoint_set('resource_manager'):
            # If management endpoint not set, use resource manager endpoint
            c.endpoints.management = c.endpoints.resource_manager
        clouds.append(c)
    for c in clouds:
        if c.name == active_cloud_name:
            c.is_active = True
            break
    return clouds


def get_cloud(cli_ctx, cloud_name):
    cloud = _get_cloud(cli_ctx, cloud_name)
    if not cloud:
        raise CloudNotRegisteredException(cloud_name)
    return cloud


def get_active_cloud(cli_ctx=None):
    if not cli_ctx:
        from azure.cli.core import get_default_cli
        cli_ctx = get_default_cli()
    try:
        cloud = get_cloud(cli_ctx, get_active_cloud_name(cli_ctx))
    except CloudNotRegisteredException as err:
        logger.warning(err)
        default_cloud_name = get_default_cloud_name()
        logger.warning("Resetting active cloud to'%s'.", default_cloud_name)
        _set_active_cloud(cli_ctx, default_cloud_name)
        cloud = get_cloud(cli_ctx, default_cloud_name)
    if cloud.profile != 'latest':
        logger.warning("Cloud profile '%s' will be deprecated starting from 2.73.0. Please use the 'latest' profile or the CLI 2.66.* (LTS) version instead.", cloud.profile)
    return cloud


def get_cloud_subscription(cloud_name):
    config = configparser.ConfigParser()
    config.read(CLOUD_CONFIG_FILE)
    try:
        return config.get(cloud_name, 'subscription')
    except (configparser.NoOptionError, configparser.NoSectionError):
        return None


def set_cloud_subscription(cli_ctx, cloud_name, subscription):
    if not _get_cloud(cli_ctx, cloud_name):
        raise CloudNotRegisteredException(cloud_name)
    config = configparser.ConfigParser()
    config.read(CLOUD_CONFIG_FILE)
    if subscription:
        try:
            config.add_section(cloud_name)
        except configparser.DuplicateSectionError:
            pass
        config.set(cloud_name, 'subscription', subscription)
    else:
        try:
            config.remove_option(cloud_name, 'subscription')
        except configparser.NoSectionError:
            pass
    if not os.path.isdir(GLOBAL_CONFIG_DIR):
        os.makedirs(GLOBAL_CONFIG_DIR)
    with open(CLOUD_CONFIG_FILE, 'w') as configfile:
        config.write(configfile)


def _set_active_subscription(cli_ctx, cloud_name):
    from azure.cli.core._profile import (Profile, _ENVIRONMENT_NAME, _SUBSCRIPTION_ID,
                                         _STATE, _SUBSCRIPTION_NAME)
    profile = Profile(cli_ctx=cli_ctx)
    subscription_to_use = get_cloud_subscription(cloud_name) or \
                          next((s[_SUBSCRIPTION_ID] for s in profile.load_cached_subscriptions()  # noqa
                                if s[_STATE] == 'Enabled'),
                               None)
    if subscription_to_use:
        try:
            profile.set_active_subscription(subscription_to_use)
            sub = profile.get_subscription(subscription_to_use)
            logger.warning("Active subscription switched to '%s (%s)'.",
                           sub[_SUBSCRIPTION_NAME], sub[_SUBSCRIPTION_ID])
        except CLIError as e:
            logger.warning(e)
            logger.warning("Unable to automatically switch the active subscription. "
                           "Use 'az account set'.")
    else:
        logger.warning("Use 'az login' to log in to this cloud.")
        logger.warning("Use 'az account set' to set the active subscription.")


def switch_active_cloud(cli_ctx, cloud_name):
    if cli_ctx.cloud.name == cloud_name:
        return
    if not _get_cloud(cli_ctx, cloud_name):
        raise CloudNotRegisteredException(cloud_name)
    _set_active_cloud(cli_ctx, cloud_name)
    logger.warning("Switched active cloud to '%s'.", cloud_name)
    _set_active_subscription(cli_ctx, cloud_name)


def _config_add_cloud(config, cloud, overwrite=False):
    """ Add a cloud to a config object """
    try:
        config.add_section(cloud.name)
    except configparser.DuplicateSectionError:
        if not overwrite:
            raise CloudAlreadyRegisteredException(cloud.name)
    if cloud.profile:
        config.set(cloud.name, 'profile', cloud.profile)
    for k, v in cloud.endpoints.__dict__.items():
        if v is not None:
            config.set(cloud.name, 'endpoint_{}'.format(k), v)
    for k, v in cloud.suffixes.__dict__.items():
        if v is not None:
            config.set(cloud.name, 'suffix_{}'.format(k), v)


def _save_cloud(cloud, overwrite=False):
    config = configparser.ConfigParser()
    config.read(CLOUD_CONFIG_FILE)
    _config_add_cloud(config, cloud, overwrite=overwrite)
    if not os.path.isdir(GLOBAL_CONFIG_DIR):
        os.makedirs(GLOBAL_CONFIG_DIR)
    with open(CLOUD_CONFIG_FILE, 'w') as configfile:
        config.write(configfile)


def add_cloud(cli_ctx, cloud):
    if _get_cloud(cli_ctx, cloud.name):
        raise CloudAlreadyRegisteredException(cloud.name)
    _save_cloud(cloud)


def update_cloud(cli_ctx, cloud):
    if not _get_cloud(cli_ctx, cloud.name):
        raise CloudNotRegisteredException(cloud.name)
    _save_cloud(cloud, overwrite=True)


def remove_cloud(cli_ctx, cloud_name):
    if not _get_cloud(cli_ctx, cloud_name):
        raise CloudNotRegisteredException(cloud_name)
    if cloud_name == cli_ctx.cloud.name:
        raise CannotUnregisterCloudException("The cloud '{}' cannot be unregistered "
                                             "as it's currently active.".format(cloud_name))
    is_known_cloud = next((x for x in KNOWN_CLOUDS if x.name == cloud_name), None)
    if is_known_cloud:
        raise CannotUnregisterCloudException("The cloud '{}' cannot be unregistered "
                                             "as it's not a custom cloud.".format(cloud_name))
    config = configparser.ConfigParser()
    config.read(CLOUD_CONFIG_FILE)
    config.remove_section(cloud_name)
    with open(CLOUD_CONFIG_FILE, 'w') as configfile:
        config.write(configfile)


def cloud_forbid_telemetry(cli_ctx):
    return get_active_cloud_name(cli_ctx) in CLOUDS_FORBIDDING_TELEMETRY



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import logging
import threading
import uuid

from azure.core.exceptions import AzureError, HttpResponseError
from azure.core.polling import NoPolling
from azure.core.polling.base_polling import LROBasePolling
from azure.core.tracing.common import with_current_context
from azure.core.tracing.decorator import distributed_trace

_LOGGER = logging.getLogger(__name__)


class AAZNoPolling(NoPolling):
    pass


class AAZBasePolling(LROBasePolling):

    def __init__(self, *args, http_response_error_callback=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._http_response_error_callback = http_response_error_callback

    def initialize(self, client, initial_response, deserialization_callback):
        try:
            super().initialize(client, initial_response, deserialization_callback)
        except HttpResponseError as err:
            if self._http_response_error_callback:
                # the HttpResponseError raise by LROBasePolling missing detailed error message
                # _http_response_error_callback will build HttpResponseError with detailed error message
                self._http_response_error_callback(err.response)
            else:
                raise err

    def run(self):
        try:
            super().run()
        except HttpResponseError as err:
            if self._http_response_error_callback:
                # the HttpResponseError raise by LROBasePolling missing detailed message
                # _http_response_error_callback will build HttpResponseError with detailed error message
                self._http_response_error_callback(err.response)
            else:
                raise err


class AAZLROPoller:

    def __init__(self, polling_generator, result_callback):
        self._callbacks = []
        self._polling_generator = polling_generator  # generator
        self._polling_method = None
        self._result_callback = result_callback

        # Prepare thread execution
        self._thread = None
        self._done = None
        self._exception = None

        self._done = threading.Event()
        self._thread = threading.Thread(
            target=with_current_context(self._start),
            name="AAZLROPoller({})".format(uuid.uuid4()))
        self._thread.daemon = True
        self._thread.start()

    def _start(self):
        """Start the long running operation.
        On completion, runs any callbacks.

        :param callable update_cmd: The API request to check the status of
         the operation.
        """
        try:
            for polling_method in self._polling_generator:
                self._polling_method = polling_method
                try:
                    self._polling_method.run()
                except AzureError as error:
                    if not error.continuation_token:
                        try:
                            error.continuation_token = self._polling_method.get_continuation_token()
                        except Exception as err:  # pylint: disable=broad-except
                            _LOGGER.warning("Unable to retrieve continuation token: %s", err)
                            error.continuation_token = None
                    raise error
        except Exception as error:  # pylint: disable=broad-except
            self._exception = error
        finally:
            self._done.set()

    def result(self, timeout=None):
        """Return the result of the long running operation, or
        the result available after the specified timeout.

        :returns: The deserialized resource of the long running operation,
         if one is available.
        :raises ~azure.core.exceptions.HttpResponseError: Server problem with the query.
        """
        self.wait(timeout)
        resource = self._polling_method.resource()
        if self._result_callback:
            return self._result_callback(resource)
        return resource

    @distributed_trace
    def wait(self, timeout=None):
        """Wait on the long running operation for a specified length
        of time. You can check if this call as ended with timeout with the
        "done()" method.

        :param float timeout: Period of time to wait for the long running
         operation to complete (in seconds).
        :raises ~azure.core.exceptions.HttpResponseError: Server problem with the query.
        """
        if self._thread is None:
            return

        self._thread.join(timeout=timeout)

        if self._exception:  # derive from BaseException
            raise self._exception

    def done(self):
        """Check status of the long running operation.

        :returns: 'True' if the process has completed, else 'False'.
        :rtype: bool
        """
        return self._thread is None or not self._thread.is_alive()



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
import abc

from collections import OrderedDict

from azure.cli.core.util import shell_safe_json_parse
from ._base import AAZBaseType, AAZValuePatch, AAZUndefined
from ._field_value import AAZObject, AAZDict, AAZFreeFormDict, AAZList, AAZSimpleValue, \
    AAZIdentityObject, AAZAnyValue
from ._utils import to_snack_case
from .exceptions import AAZUnknownFieldError, AAZConflictFieldDefinitionError, AAZValuePrecisionLossError, \
    AAZInvalidFieldError, AAZInvalidValueError

# pylint: disable=protected-access, too-few-public-methods, isinstance-second-argument-not-valid-type
# pylint: disable=too-many-instance-attributes


# build in types
class AAZSimpleType(AAZBaseType):
    """Simple value type"""

    DataType = None

    _ValueCls = AAZSimpleValue

    def process_data(self, data, **kwargs):
        if data == None:  # noqa: E711, pylint: disable=singleton-comparison
            # data can be None or AAZSimpleValue == None
            if self._nullable:
                return None
            return AAZValuePatch.build(self)

        if data == AAZUndefined:
            return AAZValuePatch.build(self)

        if isinstance(data, AAZSimpleValue):
            if data._is_patch:
                # return value patch
                return AAZValuePatch.build(self)

            data = data._data

        assert self.DataType is not None
        if not isinstance(data, self.DataType):
            try:
                json_data = shell_safe_json_parse(data)
            except:  # pylint:disable=bare-except
                raise AAZInvalidValueError('Expect {}, got {} ({})'.format(self.DataType, data, type(data)))
            if not isinstance(json_data, self.DataType):
                raise AAZInvalidValueError('Expect {}, got {} ({})'.format(self.DataType, data, type(data)))
            data = json_data

        return data


class AAZIntType(AAZSimpleType):
    DataType = int


class AAZStrType(AAZSimpleType):
    DataType = str


class AAZBoolType(AAZSimpleType):
    DataType = bool


class AAZFloatType(AAZSimpleType):
    DataType = float

    def process_data(self, data, **kwargs):
        if data == None:  # noqa: E711, pylint: disable=singleton-comparison
            # data can be None or AAZSimpleValue == None
            if self._nullable:
                return None
            return AAZValuePatch.build(self)

        if data == AAZUndefined:
            return AAZValuePatch.build(self)

        if isinstance(data, AAZSimpleValue):
            if data._is_patch:
                # return value patch
                return AAZValuePatch.build(self)

            data = data._data

        if isinstance(data, int):
            # transform int to float
            if float(data) != data:
                raise AAZValuePrecisionLossError(data, float(data))
            data = float(data)

        if not isinstance(data, self.DataType):
            try:
                json_data = shell_safe_json_parse(data)
            except:  # pylint:disable=bare-except
                raise AAZInvalidValueError('Expect {}, got {} ({})'.format(self.DataType, data, type(data)))
            if isinstance(json_data, int):
                # transform int to float
                if float(json_data) != json_data:
                    raise AAZValuePrecisionLossError(json_data, float(json_data))
                json_data = float(json_data)
            if not isinstance(json_data, self.DataType):
                raise AAZInvalidValueError('Expect {}, got {} ({})'.format(self.DataType, data, type(data)))
            data = json_data

        return data


class AAZAnyType(AAZSimpleType):
    """Any type"""

    _ValueCls = AAZAnyValue

    def process_data(self, data, **kwargs):
        if isinstance(data, (AAZObject, AAZDict, AAZList)):
            # convert them to simple dict, list, None or AAZUndefined values
            value = data.to_serialized_data()
            # TODO: may need to add warnings or raise error when treat the patch assign into the full assign.
            # for the code gen commands it will not happen as the any type args only support full assign.
            # but for the customization code, this will happen when convert AAZCompoundTypeArg to AAZAnyTypeArg
            # if data._is_patch and isinstance(value, (dict, list)) and len(value):
            #     pass
            data = value

        if data == None:  # noqa: E711, pylint: disable=singleton-comparison
            # data can be None or AAZSimpleValue == None
            if self._nullable:
                return None
            return AAZValuePatch.build(self)

        if data == AAZUndefined:
            return AAZValuePatch.build(self)

        if isinstance(data, AAZSimpleValue):
            if data._is_patch:
                # return value patch
                return AAZValuePatch.build(self)

            return data._data

        return data


# compound types
class AAZObjectType(AAZBaseType):
    """Object value type"""

    _PROTECTED_KEYWORDS = (
        "get_attr_name",
        "process_data",
        "to_serialized_data",
        "discriminate_by",
        "get_discriminator"
    )   # these keywords are used in AAZObjectValue and AAZObjectType, object should not use them as a field name.

    _ValueCls = AAZObject
    _PatchDataCls = dict

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # It's important to keep the order of fields.
        # This feature can help to resolve arguments interdependent problem.
        # aaz-dev should register fields based on the interdependent order.
        self._fields = OrderedDict()
        self._fields_alias_map = {}  # key is the option, value is field

        # Polymorphism support
        self._discriminator_field_name = None
        self._discriminators = OrderedDict()

    def __getitem__(self, key):
        name = self.get_attr_name(key)
        if name not in self._fields:
            # must raise AttributeError to support hasattr check
            raise AAZUnknownFieldError(self, key)
        return self._fields[name]

    def __setitem__(self, key, value):
        assert not key.startswith('_')
        assert key not in self._PROTECTED_KEYWORDS

        if not isinstance(value, AAZBaseType):
            raise AAZInvalidFieldError(self, key, f"unknown field type {type(value)}")

        if hasattr(self, key):
            # key should not be defined before
            raise AAZConflictFieldDefinitionError(
                self, key, "Key already been defined before")

        name = key
        value._name = name
        self._fields[name] = value

        # update alias map
        aliases = [*value._options] if value._options else []
        if value._serialized_name:
            aliases.append(value._serialized_name)

        for alias in aliases:
            if alias == name:
                continue

            assert not alias.startswith('_')
            assert alias not in self._PROTECTED_KEYWORDS

            if alias in self._fields_alias_map and self._fields_alias_map[alias] != name:
                raise AAZConflictFieldDefinitionError(
                    self, name, f"Alias is already used by other field: {self._fields_alias_map[alias]}")
            self._fields_alias_map[alias] = name

    def __getattr__(self, key):
        return self[key]

    def __setattr__(self, key, value):
        if key.startswith('_'):
            assert not isinstance(value, AAZBaseType)
            self.__dict__[key] = value
        else:
            self[key] = value

    def get_attr_name(self, key):
        if key in self._fields:
            return key
        if key in self._fields_alias_map:
            return self._fields_alias_map[key]
        if isinstance(key, str) and key != key.lower():
            # if key is not found convert camel case key to snack case key
            key = to_snack_case(key)
            if key in self._fields:
                return key
        return None

    def process_data(self, data, **kwargs):
        if data == None:  # noqa: E711, pylint: disable=singleton-comparison
            # data can be None or AAZSimpleValue == None
            if self._nullable:
                return None
            return AAZValuePatch.build(self)

        if data == AAZUndefined:
            return AAZValuePatch.build(self)

        if isinstance(data, AAZObject) and data._is_patch:
            # use value patch
            result = AAZValuePatch.build(self)
        else:
            result = {}
        value = AAZObject(schema=self, data=result)

        if isinstance(data, AAZObject):
            if self._discriminator_field_name:
                # assign discriminator field first
                for key in data._data.keys():
                    name = self.get_attr_name(key)
                    if name == self._discriminator_field_name:
                        value[name] = data[key]
                        break

            for key in data._data.keys():
                if not hasattr(value, key):
                    # ignore undefined key
                    continue
                value[key] = data[key]

        else:
            if not isinstance(data, dict):
                raise AAZInvalidValueError("Expect <class 'dict'>, got {} ({})".format(data, type(data)))

            if self._discriminator_field_name:
                # assign discriminator field first
                for key, sub_data in data.items():
                    name = self.get_attr_name(key)
                    if name == self._discriminator_field_name:
                        value[name] = sub_data
                        break

            for key, sub_data in data.items():
                if not hasattr(value, key):
                    # ignore undefined key
                    continue
                value[key] = sub_data

        return result

    # Polymorphism support
    def discriminate_by(self, key, data, schema=None):

        name = self.get_attr_name(key)
        if name not in self._fields:
            raise AAZUnknownFieldError(self, key)
        field = self._fields[name]
        if not isinstance(field, AAZStrType):
            raise AAZInvalidFieldError(self, name, f"Invalid discriminator field type: {type(field)}")
        data = field.process_data(data)
        if self._discriminator_field_name is None:
            self._discriminator_field_name = name
        elif self._discriminator_field_name != name:
            raise AAZConflictFieldDefinitionError(
                self, name, f"Conflict discriminator field name with: {self._discriminator_field_name}")
        schema = schema or AAZObjectType()  # use provided schema or create a object type
        assert isinstance(schema, AAZObjectType)
        return self._discriminators.setdefault(data, schema)

    def get_discriminator(self, data):
        if self._discriminator_field_name is None:
            return None
        if data == AAZUndefined or not data:
            return None
        if isinstance(data, AAZObject):
            data = data._data
        if not isinstance(data, dict):
            raise AAZInvalidValueError("Expect <class 'dict'>, got {} ({})".format(data, type(data)))

        for key, field_data in data.items():
            name = self.get_attr_name(key)
            if name == self._discriminator_field_name:
                return self._discriminators.get(field_data, None)
        return None


class AAZBaseDictType(AAZBaseType):

    _PatchDataCls = dict

    @abc.abstractmethod
    def __getitem__(self, key):
        raise NotImplementedError()

    def process_data(self, data, **kwargs):
        if data == None:  # noqa: E711, pylint: disable=singleton-comparison
            # data can be None or AAZSimpleValue == None
            if self._nullable:
                return None
            return AAZValuePatch.build(self)

        if data == AAZUndefined:
            return AAZValuePatch.build(self)

        if isinstance(data, self._ValueCls) and data._is_patch:
            # use value patch
            result = AAZValuePatch.build(self)
        else:
            result = {}

        value = self._ValueCls(schema=self, data=result)  # pylint: disable=not-callable

        if isinstance(data, self._ValueCls):
            for key in data._data.keys():
                value[key] = data[key]
        else:
            if not isinstance(data, dict):
                raise AAZInvalidValueError("Expect <class 'dict'>, got {} ({})".format(data, type(data)))

            for key, sub_data in data.items():
                value[key] = sub_data
        return result


class AAZDictType(AAZBaseDictType):
    """Dict value type"""

    _ValueCls = AAZDict

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._element = None

    @property
    def Element(self):
        if self._element is None:
            raise AAZUnknownFieldError(self, "Element")
        return self._element

    @Element.setter
    def Element(self, value):
        if self._element is None:
            assert isinstance(value, AAZBaseType)
            self._element = value
            assert self._element._name is None
            assert not self._element._options
            assert self._element._serialized_name is None
        elif self._element != value:
            raise AAZConflictFieldDefinitionError(self, "Element", "Redefine element in different schema")

    def __getitem__(self, key):
        return self.Element


# Warning: This type should not be used any more, the new aaz-dev-tools only use AAZDictType with AAZAnyType
class AAZFreeFormDictType(AAZDictType):
    """Free form dict value type"""

    _ValueCls = AAZFreeFormDict

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # for backward compatible, so support nullable for any type.
        # from the new code gen tools, it will avoid using AAZFreeFormDictType
        self._element = AAZAnyType(nullable=True)


class AAZListType(AAZBaseType):
    """List value type"""

    _ValueCls = AAZList
    _PatchDataCls = dict

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._element = None

    @property
    def Element(self):
        if self._element is None:
            raise AAZUnknownFieldError(self, "Element")
        return self._element

    @Element.setter
    def Element(self, value):
        if self._element is None:
            assert isinstance(value, AAZBaseType)
            self._element = value
            assert self._element._name is None
            assert not self._element._options
            assert self._element._serialized_name is None
        elif self._element != value:
            raise AAZConflictFieldDefinitionError(self, "Element", "Redefine element in different schema")

    def __getitem__(self, key):
        return self.Element

    def process_data(self, data, **kwargs):
        if data == None:  # noqa: E711, pylint: disable=singleton-comparison
            # data can be None or AAZSimpleValue == None
            if self._nullable:
                return None
            return AAZValuePatch.build(self)

        if data == AAZUndefined:
            return AAZValuePatch.build(self)

        if isinstance(data, AAZList) and data._is_patch:
            # use value patch
            result = AAZValuePatch.build(self)
        else:
            result = {}
        value = AAZList(schema=self, data=result)

        if isinstance(data, AAZList):
            for idx in data._data.keys():
                value[idx] = data[idx]
        else:
            if not isinstance(data, list):
                raise AAZInvalidValueError("Expect <class 'list'>, got {} ({})".format(data, type(data)))

            for idx, sub_data in enumerate(data):
                value[idx] = sub_data

        return result


class AAZIdentityObjectType(AAZObjectType):
    _ValueCls = AAZIdentityObject



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
Credentials to acquire tokens from MSAL.
"""

from knack.log import get_logger
from knack.util import CLIError
from msal import (PublicClientApplication, ConfidentialClientApplication,
                  ManagedIdentityClient, SystemAssignedManagedIdentity, UserAssignedManagedIdentity)

from .constants import AZURE_CLI_CLIENT_ID
from .util import check_result

logger = get_logger(__name__)


class UserCredential:  # pylint: disable=too-few-public-methods

    def __init__(self, client_id, username, **kwargs):
        """User credential wrapping msal.application.PublicClientApplication

        :param client_id: Client ID of the CLI.
        :param username: The username for user credential.
        """
        self._msal_app = PublicClientApplication(client_id, **kwargs)

        # Make sure username is specified, otherwise MSAL returns all accounts
        assert username, "username must be specified, got {!r}".format(username)

        accounts = self._msal_app.get_accounts(username)

        # Usernames are usually unique. We are collecting corner cases to better understand its behavior.
        if len(accounts) > 1:
            raise CLIError(f"Found multiple accounts with the same username '{username}': {accounts}\n"
                           "Please report to us via Github: https://github.com/Azure/azure-cli/issues/20168")

        if not accounts:
            raise CLIError("User '{}' does not exist in MSAL token cache. Run `az login`.".format(username))

        self._account = accounts[0]

    def acquire_token(self, scopes, claims_challenge=None, **kwargs):
        # scopes must be a list.
        # For acquiring SSH certificate, scopes is ['https://pas.windows.net/CheckMyAccess/Linux/.default']
        # kwargs is already sanitized by CredentialAdaptor, so it can be safely passed to MSAL
        logger.debug("UserCredential.acquire_token: scopes=%r, claims_challenge=%r, kwargs=%r",
                     scopes, claims_challenge, kwargs)

        # Apply agentic session parameters for user identity flows
        from .agentic_session import build_agentic_session_params, merge_access_token_claims
        agentic_session_id, agentic_claims = build_agentic_session_params()
        if agentic_session_id:
            # Both paths: client_session in data and params so eSTS can identify the agentic session
            kwargs["data"] = kwargs.get("data") or {}
            kwargs["data"]["client_session"] = agentic_session_id
            kwargs["params"] = kwargs.get("params") or {}
            kwargs["params"]["client_session"] = agentic_session_id

            if getattr(self._msal_app, '_enable_broker', False):
                # Broker path: claims_challenge flows to MSALRuntime cache key via set_decoded_claims.
                # This causes MSAL to skip its local AT cache and forward claims to the broker,
                # where requestedClaims becomes part of the C++ cache key.
                claims_challenge = merge_access_token_claims(claims_challenge, agentic_claims)
            # Non-broker path: client_session in data flows into ext_cache_key (SHA256 hash),
            # which partitions the MSAL Python token cache. No claims_challenge needed.

            from azure.cli.core.telemetry import set_agentic_session
            set_agentic_session(True)

        if claims_challenge:
            logger.info('Acquiring new access token silently with claims challenge: %s', claims_challenge)
        result = self._msal_app.acquire_token_silent_with_error(
            scopes, self._account, claims_challenge=claims_challenge, **kwargs)

        from azure.cli.core.azclierror import AuthenticationError
        try:
            # Check if an access token is returned.
            check_result(result, tenant=self._msal_app.authority.tenant, scopes=scopes,
                         claims_challenge=claims_challenge)
        except AuthenticationError as ex:
            # For VM SSH ('data' is passed), if getting access token fails because
            # Conditional Access MFA step-up or compliance check is required, re-launch
            # web browser and do auth code flow again.
            # We assume the `az ssh` command is run on a system with GUI where a web
            # browser is available.
            if 'data' in kwargs:
                logger.warning(ex)
                result = self._acquire_token_interactive(scopes, **kwargs)
            # For other scenarios like Storage Conditional Access MFA step-up, do not
            # launch browser, but show the error message and `az login` command instead.
            else:
                raise
        return result

    def _acquire_token_interactive(self, scopes, **kwargs):
        from .util import read_response_templates
        success_template, error_template = read_response_templates()

        def _prompt_launching_ui(ui=None, **_):
            logger.warning(
                "Interactively acquiring token for scope '%s'. Continue the login in the %s.",
                ' '.join(scopes), 'web browser' if ui == 'browser' else 'pop-up window')

        result = self._msal_app.acquire_token_interactive(
            scopes, login_hint=self._account['username'],
            port=8400 if self._msal_app.authority.is_adfs else None,
            success_template=success_template, error_template=error_template,
            parent_window_handle=self._msal_app.CONSOLE_WINDOW_HANDLE,
            on_before_launching_ui=_prompt_launching_ui,
            **kwargs)
        check_result(result)
        return result


class ServicePrincipalCredential:  # pylint: disable=too-few-public-methods

    def __init__(self, client_id, client_credential, **kwargs):
        """Service principal credential wrapping msal.application.ConfidentialClientApplication.

        :param client_id: The service principal's client ID.
        :param client_credential: client_credential that will be passed to MSAL.
        """
        self._msal_app = ConfidentialClientApplication(client_id, client_credential=client_credential, **kwargs)

    def acquire_token(self, scopes, **kwargs):
        logger.debug("ServicePrincipalCredential.acquire_token: scopes=%r, kwargs=%r", scopes, kwargs)
        result = self._msal_app.acquire_token_for_client(scopes, **kwargs)
        check_result(result)
        return result


class CloudShellCredential:  # pylint: disable=too-few-public-methods
    # Cloud Shell acts as a "broker" to obtain access token for the user account, so even though it uses
    # managed identity protocol, it returns a user token.
    # That's why MSAL uses acquire_token_interactive to retrieve an access token in Cloud Shell.
    # See https://github.com/Azure/azure-cli/pull/29637

    def __init__(self):
        self._msal_app = PublicClientApplication(
            AZURE_CLI_CLIENT_ID,  # Use a real client_id, so that cache would work
            # TODO: We currently don't maintain an MSAL token cache as Cloud Shell already has its own token cache.
            #   Ideally we should also use an MSAL token cache.
            #   token_cache=...
        )

    def acquire_token(self, scopes, **kwargs):
        logger.debug("CloudShellCredential.acquire_token: scopes=%r, kwargs=%r", scopes, kwargs)
        result = self._msal_app.acquire_token_interactive(scopes, prompt="none", **kwargs)
        check_result(result, scopes=scopes)
        return result


class ManagedIdentityCredential:  # pylint: disable=too-few-public-methods
    """Managed identity credential implementing get_token interface.
    Currently, only Azure Arc's system-assigned managed identity is supported.
    """

    def __init__(self, client_id=None, resource_id=None, object_id=None):
        import requests
        if client_id or resource_id or object_id:
            managed_identity = UserAssignedManagedIdentity(
                client_id=client_id, resource_id=resource_id, object_id=object_id)
        else:
            managed_identity = SystemAssignedManagedIdentity()
        self._msal_client = ManagedIdentityClient(managed_identity, http_client=requests.Session())

    def acquire_token(self, scopes, **kwargs):
        logger.debug("ManagedIdentityCredential.acquire_token: scopes=%r, kwargs=%r", scopes, kwargs)

        from .util import scopes_to_resource
        result = self._msal_client.acquire_token_for_client(resource=scopes_to_resource(scopes))
        check_result(result)
        return result



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
Support styled output.

Currently, only color is supported, underline/bold/italic may be supported in the future.

Design spec:
https://devdivdesignguide.azurewebsites.net/command-line-interface/color-guidelines-for-command-line-interface/

Console Virtual Terminal Sequences:
https://learn.microsoft.com/en-us/windows/console/console-virtual-terminal-sequences#text-formatting
"""

import sys
from enum import Enum

from knack.log import get_logger
from knack.util import is_modern_terminal


logger = get_logger(__name__)


class Style(str, Enum):
    PRIMARY = "primary"
    SECONDARY = "secondary"
    HIGHLIGHT = "highlight"
    IMPORTANT = "important"
    ACTION = "action"  # name TBD
    HYPERLINK = "hyperlink"
    # Message colors
    ERROR = "error"
    SUCCESS = "success"
    WARNING = "warning"


def _rgb_hex(rgb_hex: str):
    """
    Convert RGB hex value to Control Sequences.
    """
    template = '\x1b[38;2;{r};{g};{b}m'
    if rgb_hex.startswith("#"):
        rgb_hex = rgb_hex[1:]

    rgb = {}
    for i, c in enumerate(('r', 'g', 'b')):
        value_str = rgb_hex[i * 2: i * 2 + 2]
        value_int = int(value_str, 16)
        rgb[c] = value_int

    return template.format(**rgb)


DEFAULT = '\x1b[0m'  # Default

# Theme that doesn't contain any style
THEME_NONE = None

# Theme to be used in a dark-themed terminal
THEME_DARK = {
    Style.PRIMARY: DEFAULT,
    Style.SECONDARY: '\x1b[90m',  # Bright Foreground Black
    Style.HIGHLIGHT: '\x1b[96m',  # Bright Foreground Cyan
    Style.IMPORTANT: '\x1b[95m',  # Bright Foreground Magenta
    Style.ACTION: '\x1b[94m',  # Bright Foreground Blue
    Style.HYPERLINK: '\x1b[96m',  # Bright Foreground Cyan
    Style.ERROR: '\x1b[91m',  # Bright Foreground Red
    Style.SUCCESS: '\x1b[92m',  # Bright Foreground Green
    Style.WARNING: '\x1b[93m'  # Bright Foreground Yellow
}

# Theme to be used in a light-themed terminal
THEME_LIGHT = {
    Style.PRIMARY: DEFAULT,
    Style.SECONDARY: '\x1b[90m',  # Bright Foreground Black
    Style.HIGHLIGHT: '\x1b[36m',  # Foreground Cyan
    Style.IMPORTANT: '\x1b[35m',  # Foreground Magenta
    Style.ACTION: '\x1b[34m',  # Foreground Blue
    Style.HYPERLINK: '\x1b[36m',  # Foreground Cyan
    Style.ERROR: '\x1b[31m',  # Foreground Red
    Style.SUCCESS: '\x1b[32m',  # Foreground Green
    Style.WARNING: '\x1b[33m'  # Foreground Yellow
}

# Theme to be used in Cloud Shell
# Text and background's Contrast Ratio should be above 4.5:1
THEME_CLOUD_SHELL = {
    Style.PRIMARY: _rgb_hex('#ffffff'),
    Style.SECONDARY: _rgb_hex('#bcbcbc'),
    Style.HIGHLIGHT: _rgb_hex('#72d7d8'),  # Same as Style.HYPERLINK, for now
    Style.IMPORTANT: _rgb_hex('#f887ff'),
    Style.ACTION: _rgb_hex('#6cb0ff'),
    Style.HYPERLINK: _rgb_hex('#72d7d8'),
    Style.ERROR: _rgb_hex('#f55d5c'),
    Style.SUCCESS: _rgb_hex('#70d784'),
    Style.WARNING: _rgb_hex('#fbd682'),
}


class Theme(str, Enum):
    DARK = 'dark'
    LIGHT = 'light'
    CLOUD_SHELL = 'cloud-shell'
    NONE = 'none'


THEME_DEFINITIONS = {
    Theme.DARK: THEME_DARK,
    Theme.LIGHT: THEME_LIGHT,
    Theme.CLOUD_SHELL: THEME_CLOUD_SHELL,
    Theme.NONE: THEME_NONE
}

# Blue and bright blue is not visible under the default theme of powershell.exe
POWERSHELL_COLOR_REPLACEMENT = {
    '\x1b[34m': DEFAULT,  # Foreground Blue
    '\x1b[94m': DEFAULT  # Bright Foreground Blue
}


def print_styled_text(*styled_text_objects, file=None, **kwargs):
    """
    Print styled text. This function wraps the built-in function `print`, additional arguments can be sent
    via keyword arguments.

    :param styled_text_objects: The input text objects. See format_styled_text for formats of each object.
    :param file: The file to print the styled text. The default target is sys.stderr.
    """
    formatted_list = [format_styled_text(obj) for obj in styled_text_objects]
    # Always fetch the latest sys.stderr in case it has been wrapped by colorama.
    print(*formatted_list, file=file or sys.stderr, **kwargs)


def format_styled_text(styled_text, theme=None):
    """Format styled text. Dark theme used by default. Available themes are 'dark', 'light', 'none'.

    To change theme for all invocations of this function, set `format_styled_text.theme`.
    To change theme for one invocation, set parameter `theme`.

    :param styled_text: Can be in these formats:
        - text
        - (style, text)
        - [(style, text), ...]
    :param theme: The theme used to format text. Can be theme name str, `Theme` Enum or dict.
    """
    if theme is None:
        theme = getattr(format_styled_text, "theme", THEME_DARK)

    # Convert str to the theme dict
    if isinstance(theme, str):
        theme = get_theme_dict(theme)

    # If style is enabled, cache the value of is_legacy_powershell.
    # Otherwise if theme is None, is_legacy_powershell is meaningless.
    is_legacy_powershell = None
    if theme:
        if not hasattr(format_styled_text, "_is_legacy_powershell"):
            from azure.cli.core.util import get_parent_proc_name
            is_legacy_powershell = not is_modern_terminal() and get_parent_proc_name() == "powershell.exe"
            setattr(format_styled_text, "_is_legacy_powershell", is_legacy_powershell)
        is_legacy_powershell = getattr(format_styled_text, "_is_legacy_powershell")

    # https://python-prompt-toolkit.readthedocs.io/en/stable/pages/printing_text.html#style-text-tuples
    formatted_parts = []

    # A str as PRIMARY text
    if isinstance(styled_text, str):
        styled_text = [(Style.PRIMARY, styled_text)]

    # A tuple
    if isinstance(styled_text, tuple):
        styled_text = [styled_text]

    for text in styled_text:
        # str can also be indexed, bypassing IndexError, so explicitly check if the type is tuple
        if not (isinstance(text, tuple) and len(text) == 2):
            from azure.cli.core.azclierror import CLIInternalError
            raise CLIInternalError("Invalid styled text. It should be a list of 2-element tuples.")

        style, raw_text = text

        if theme:
            try:
                escape_seq = theme[style]
            except KeyError:
                if style.startswith('\x1b['):
                    escape_seq = style
                else:
                    from azure.cli.core.azclierror import CLIInternalError
                    raise CLIInternalError("Invalid style. Please use pre-defined style in Style enum "
                                           "or give a valid ANSI code.")
            # Replace blue in powershell.exe
            if is_legacy_powershell and escape_seq in POWERSHELL_COLOR_REPLACEMENT:
                escape_seq = POWERSHELL_COLOR_REPLACEMENT[escape_seq]
            formatted_parts.append(escape_seq + raw_text)
        else:
            formatted_parts.append(raw_text)

    # Reset control sequence
    if theme is not THEME_NONE:
        formatted_parts.append(DEFAULT)
    return ''.join(formatted_parts)


def highlight_command(raw_command):
    """Highlight a command with colors.

    For example, for

        az group create --name myrg --location westus

    The command name 'az group create', argument name '--name', '--location' are marked as ACTION style.
    The argument value 'myrg' and 'westus' are marked as PRIMARY style.
    If the argument is provided as '--location=westus', it will be marked as PRIMARY style.

    :param raw_command: The command that needs to be highlighted.
    :type raw_command: str
    :return: The styled command text.
    :rtype: list
    """

    styled_command = []
    argument_begins = False

    for index, arg in enumerate(raw_command.split()):
        spaced_arg = ' {}'.format(arg) if index > 0 else arg
        style = Style.PRIMARY

        if arg.startswith('-') and '=' not in arg:
            style = Style.ACTION
            argument_begins = True
        elif not argument_begins and '=' not in arg:
            style = Style.ACTION

        styled_command.append((style, spaced_arg))

    return styled_command


def get_theme_dict(theme: str):
    try:
        return THEME_DEFINITIONS[theme]
    except KeyError as ex:
        available_themes = ', '.join([m.value for m in Theme.__members__.values()])  # pylint: disable=no-member
        logger.warning("Invalid theme %s. Supported themes: %s", ex, available_themes)
        return None



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import difflib

import argparse
import argcomplete

from azure.cli.core import telemetry
from azure.cli.core.extension import get_extension
from azure.cli.core.commands import ExtensionCommandSource
from azure.cli.core.commands import AzCliCommandInvoker
from azure.cli.core.commands.events import EVENT_INVOKER_ON_TAB_COMPLETION
from azure.cli.core.command_recommender import CommandRecommender
from azure.cli.core.azclierror import UnrecognizedArgumentError
from azure.cli.core.azclierror import RequiredArgumentMissingError
from azure.cli.core.azclierror import InvalidArgumentValueError
from azure.cli.core.azclierror import ArgumentUsageError
from azure.cli.core.azclierror import CommandNotFoundError
from azure.cli.core.azclierror import ValidationError

from knack.log import get_logger
from knack.parser import CLICommandParser
from knack.util import CLIError

logger = get_logger(__name__)

EXTENSION_REFERENCE = ("If the command is from an extension, "
                       "please make sure the corresponding extension is installed. "
                       "To learn more about extensions, please visit "
                       "'https://learn.microsoft.com/cli/azure/azure-cli-extensions-overview'")

OVERVIEW_REFERENCE = "https://aka.ms/cli_ref"


class IncorrectUsageError(CLIError):
    '''Raised when a command is incorrectly used and the usage should be
    displayed to the user.
    '''
    pass  # pylint: disable=unnecessary-pass


class AzCompletionFinder(argcomplete.CompletionFinder):

    def _get_completions(self, comp_words, cword_prefix, cword_prequote, last_wordbreak_pos):
        external_completions = []
        self._parser.cli_ctx.raise_event(EVENT_INVOKER_ON_TAB_COMPLETION,
                                         external_completions=external_completions,
                                         parser=self._parser,
                                         comp_words=comp_words,
                                         cword_prefix=cword_prefix,
                                         cword_prequote=cword_prequote,
                                         last_wordbreak_pos=last_wordbreak_pos)

        return external_completions + super()._get_completions(comp_words,
                                                               cword_prefix,
                                                               cword_prequote,
                                                               last_wordbreak_pos)


class AzCliCommandParser(CLICommandParser):
    """ArgumentParser implementation specialized for the Azure CLI utility."""

    def __init__(self, cli_ctx=None, cli_help=None, **kwargs):
        self.command_source = kwargs.pop('_command_source', None)
        self._raw_arguments = None
        self._namespace = None
        self._suggestion_msg = []
        self.subparser_map = {}
        self.specified_arguments = []
        super().__init__(cli_ctx, cli_help=cli_help, **kwargs)

    def load_command_table(self, command_loader):
        """Load a command table into our parser."""
        # If we haven't already added a subparser, we
        # better do it.
        cmd_tbl = command_loader.command_table
        grp_tbl = command_loader.command_group_table
        if not self.subparsers:
            sp = self.add_subparsers(dest='_command_package')
            sp.required = True
            self.subparsers = {(): sp}

        for command_name, metadata in cmd_tbl.items():
            subparser = self._get_subparser(command_name.split(), grp_tbl)
            deprecate_info = metadata.deprecate_info
            if not subparser or (deprecate_info and deprecate_info.expired()):
                continue

            command_verb = command_name.split()[-1]

            # inject command_module designer's help formatter -- default is HelpFormatter
            fc = metadata.formatter_class or argparse.HelpFormatter

            command_parser = subparser.add_parser(command_verb,
                                                  description=metadata.description,
                                                  parents=self.parents,
                                                  conflict_handler='error',
                                                  help_file=metadata.help,
                                                  formatter_class=fc,
                                                  cli_help=self.cli_help,
                                                  _command_source=metadata.command_source)
            self.subparser_map[command_name] = command_parser
            command_parser.cli_ctx = self.cli_ctx
            command_validator = metadata.validator
            argument_validators = []
            argument_groups = {}
            for _, arg in metadata.arguments.items():
                # don't add deprecated arguments to the parser
                deprecate_info = arg.type.settings.get('deprecate_info', None)
                if deprecate_info and deprecate_info.expired():
                    continue

                if arg.validator:
                    argument_validators.append(arg.validator)
                try:
                    if arg.arg_group:
                        try:
                            group = argument_groups[arg.arg_group]
                        except KeyError:
                            # group not found so create
                            group_name = '{} Arguments'.format(arg.arg_group)
                            group = command_parser.add_argument_group(arg.arg_group, group_name)
                            argument_groups[arg.arg_group] = group
                        param = AzCliCommandParser._add_argument(group, arg)
                    else:
                        param = AzCliCommandParser._add_argument(command_parser, arg)
                except argparse.ArgumentError as ex:
                    raise CLIError("command authoring error for '{}': '{}' {}".format(
                        command_name, ex.args[0].dest, ex.message))  # pylint: disable=no-member
                param.completer = arg.completer
                param.deprecate_info = arg.deprecate_info
                param.preview_info = arg.preview_info
                param.experimental_info = arg.experimental_info
                param.default_value_source = arg.default_value_source
            command_parser.set_defaults(
                func=metadata,
                command=command_name,
                _cmd=metadata,
                _command_validator=command_validator,
                _argument_validators=argument_validators,
                _parser=command_parser)

    def validation_error(self, message):
        az_error = ValidationError(message)
        az_error.print_error()
        az_error.send_telemetry()
        self.exit(2)

    def error(self, message):
        # Get a recommended command from the CommandRecommender
        command_arguments = self._get_failure_recovery_arguments()
        cli_ctx = self.cli_ctx or (self.cli_help.cli_ctx if self.cli_help else None)
        recommender = CommandRecommender(*command_arguments, message, cli_ctx)
        recommender.set_help_examples(self.get_examples(self.prog))
        recommendations = recommender.provide_recommendations()

        az_error = ArgumentUsageError(message)
        if 'unrecognized arguments' in message:
            az_error = UnrecognizedArgumentError(message)
        elif 'arguments are required' in message:
            az_error = RequiredArgumentMissingError(message)
        elif 'invalid' in message:
            az_error = InvalidArgumentValueError(message)

        if '--query' in message:
            from azure.cli.core.util import QUERY_REFERENCE
            az_error.set_recommendation(QUERY_REFERENCE)
        elif recommendations:
            az_error.set_example_recommendation(recommendations)
        az_error.print_error()
        az_error.send_telemetry()
        self.exit(2)

    def format_help(self):
        extension_version = None
        extension_name = None
        try:
            if isinstance(self.command_source, ExtensionCommandSource):
                extension_name = self.command_source.extension_name
                extension_version = get_extension(self.command_source.extension_name).version
        except Exception:  # pylint: disable=broad-except
            pass

        telemetry.set_command_details(
            command=self.prog[3:],
            parameters=['--help'],
            extension_name=extension_name,
            extension_version=extension_version)
        telemetry.set_success(summary='show help')
        super().format_help()

    def get_examples(self, command):
        if not self.cli_help:
            return []
        is_group = self.is_group()
        return self.cli_help.get_examples(command,
                                          self._actions[-1] if is_group else self,
                                          is_group)

    def enable_autocomplete(self):
        argcomplete.autocomplete = AzCompletionFinder()
        argcomplete.autocomplete(self, validator=lambda c, p: c.lower().startswith(p.lower()),
                                 default_completer=lambda *args, **kwargs: ())

    def _get_failure_recovery_arguments(self, action=None):
        # Strip the leading "az " and any extraneous whitespace.
        command = self.prog[3:].strip()
        parameters = []
        parameter_set = set()
        raw_arguments = None
        extension = None

        # Extract only parameter names to ensure GPDR compliance
        def extract_safe_params(parameters):
            return AzCliCommandInvoker._extract_parameter_names(parameters)  # pylint: disable=protected-access

        # Check for extension name attribute
        def has_extension_name(command_source):
            is_extension_command_source = isinstance(command_source, ExtensionCommandSource)
            has_extension_name = False

            if is_extension_command_source:
                has_extension_name = hasattr(command_source, 'extension_name')

            return is_extension_command_source and has_extension_name

        # If the arguments have been processed into a namespace...
        if self._namespace:
            # Select the parsed command.
            if hasattr(self._namespace, 'command'):
                command = self._namespace.command
        # Parse parameter names from user input.
        if self._raw_arguments:
            raw_arguments = self._raw_arguments
            parameters = extract_safe_params(self._raw_arguments)

        for parameter in parameters:
            parameter_set.add(parameter)

        # If we can retrieve the extension from the current parser's command source...
        if has_extension_name(self.command_source):
            extension = self.command_source.extension_name
        # Otherwise, the command may have not been in a command group. The command source will not be
        # set in this case.
        elif action and action.dest in ('_subcommand', '_command_package'):
            # Get all parsers in the set of possible actions.
            parsers = list(action.choices.values())
            parser = parsers[0] if parsers else None
            # If the first parser comes from an extension...
            if parser and has_extension_name(parser.command_source):
                # We're looking for a subcommand under an extension command group. Set the
                # extension to reflect this.
                extension = parser.command_source.extension_name
            # Extend the command if the first raw argument is not a parameter.
            if raw_arguments and raw_arguments[0] not in parameter_set:
                command = '{cmd} {arg}'.format(cmd=command, arg=raw_arguments[0])
        # Otherwise, only set the extension if every subparser comes from an extension. This occurs
        # when an unrecognized argument is passed to a command from an extension.
        elif isinstance(self.subparser_map, dict):
            for _, subparser in self.subparser_map.items():
                if isinstance(subparser.command_source, ExtensionCommandSource):
                    extension = subparser.command_source.extension_name
                else:
                    extension = None
                    break

        return command, self._raw_arguments, extension

    def _get_values(self, action, arg_strings):
        value = super()._get_values(action, arg_strings)
        if action.dest and isinstance(action.dest, str) and not action.dest.startswith('_'):
            self.specified_arguments.append(action.dest)
        return value

    def parse_known_args(self, args=None, namespace=None):
        # retrieve the raw argument list in case parsing known arguments fails.
        self._raw_arguments = args
        # if parsing known arguments succeeds, get the command namespace and the argument list
        self._namespace, self._raw_arguments = super().parse_known_args(args=args, namespace=namespace)
        return self._namespace, self._raw_arguments

    def _check_value(self, action, value):
        # Override to customize the error message when a argument is not among the available choices
        # converted value must be one of the choices (if specified)
        if action.choices is not None and value not in action.choices:  # pylint: disable=too-many-nested-blocks
            # self.cli_ctx is None when self.prog is beyond 'az', such as 'az iot'.
            # use cli_ctx from cli_help which is not lost.
            cli_ctx = self.cli_ctx or (self.cli_help.cli_ctx if self.cli_help else None)

            command_name_inferred = self.prog
            use_dynamic_install = 'no'
            if not self.command_source:
                from azure.cli.core.extension.dynamic_install import try_install_extension
                candidates = []
                args = self.prog.split() + self._raw_arguments
                # Check if the command is from an extension. If yes, try to fix by installing the extension, then exit.
                # The command will be rerun in another process.
                use_dynamic_install = try_install_extension(self, args)
                # parser has no `command_source`, value is part of command itself
                error_msg = "'{value}' is misspelled or not recognized by the system.".format(value=value)
                az_error = CommandNotFoundError(error_msg)
                candidates = difflib.get_close_matches(value, action.choices, cutoff=0.7)
                if candidates:
                    # use the most likely candidate to replace the misspelled command
                    args_inferred = [item if item != value else candidates[0] for item in args]
                    command_name_inferred = ' '.join(args_inferred).split('-', maxsplit=1)[0]
            else:
                # `command_source` indicates command values have been parsed, value is an argument
                parameter = action.option_strings[0] if action.option_strings else action.dest
                error_msg = "{prog}: '{value}' is not a valid value for '{param}'. Allowed values: {choices}.".format(
                    prog=self.prog, value=value, param=parameter, choices=', '.join([str(x) for x in action.choices]))
                az_error = InvalidArgumentValueError(error_msg)
                candidates = difflib.get_close_matches(value, action.choices, cutoff=0.7)

            command_arguments = self._get_failure_recovery_arguments(action)
            if candidates:
                az_error.set_recommendation("Did you mean '{}' ?".format(candidates[0]))

            # recommend a command for user
            recommender = CommandRecommender(*command_arguments, error_msg, cli_ctx)
            recommender.set_help_examples(self.get_examples(command_name_inferred))
            recommendations = recommender.provide_recommendations()
            if recommendations:
                az_error.set_example_recommendation(recommendations)

            # remind user to check extensions if we can not find a command to recommend
            if isinstance(az_error, CommandNotFoundError) \
                    and not az_error.recommendations and self.prog == 'az' \
                    and use_dynamic_install == 'no':
                az_error.set_recommendation(EXTENSION_REFERENCE)

            az_error.print_error()
            az_error.send_telemetry()

            self.exit(2)
